__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.ast import base_statement, sprite_container


class AltSpritesBlock(base_statement.BaseStatement):
    """
    AST Node for alternative graphics. These are normally 32bpp graphics, possible
    for a higher zoom-level than the default sprites.
    Syntax: alternative_sprites(name, zoom_level, bit_depth[, image_file])

    @ivar name: The name of the replace/font_glyph/replace_new/spriteset/base_graphics-block this
                block contains alternative graphics for.
    @type name: L{expression.Identifier}

    @ivar zoom_level: The zoomlevel these graphics are for.
    @type zoom_level: C{int}

    @ivar bit_depth: Bit depth these graphics are for
    @type bit_depth: C{int}

    @ivar image_file: Default graphics file for the sprites in this block.
    @type image_file: L{expression.StringLiteral} or C{None}

    @ivar mask_file: Default graphics file for the mask sprites in this block.
    @type mask_file: L{expression.StringLiteral} or C{None}

    @ivar sprite_list: List of real sprites or templates expanding to real sprites.
    @type sprite_list: Heterogeneous C{list} of L{RealSprite}, L{TemplateUsage}
    """

    def __init__(self, param_list, sprite_list, pos):
        base_statement.BaseStatement.__init__(self, "alt_sprites-block", pos)
        if not (3 <= len(param_list) <= 5):
            raise generic.ScriptError(
                "alternative_sprites-block requires 3 or 4 parameters, encountered " + str(len(param_list)), pos
            )

        self.name = param_list[0]
        if not isinstance(self.name, expression.Identifier):
            raise generic.ScriptError("alternative_sprites parameter 1 'name' must be an identifier", self.name.pos)

        if isinstance(param_list[1], expression.Identifier) and param_list[1].value in global_constants.zoom_levels:
            self.zoom_level = global_constants.zoom_levels[param_list[1].value]
        else:
            raise generic.ScriptError(
                "value for alternative_sprites parameter 2 'zoom level' is not a valid zoom level", param_list[1].pos
            )

        if isinstance(param_list[2], expression.Identifier) and param_list[2].value in global_constants.bit_depths:
            self.bit_depth = global_constants.bit_depths[param_list[2].value]
        else:
            raise generic.ScriptError(
                "value for alternative_sprites parameter 3 'bit depth' is not a valid bit depth", param_list[2].pos
            )
        if self.bit_depth == 32:
            global_constants.any_32bpp_sprites = global_constants.allow_32bpp

        if len(param_list) >= 4:
            self.image_file = param_list[3].reduce()
            if not isinstance(self.image_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "alternative_sprites-block parameter 4 'file' must be a string literal", self.image_file.pos
                )
        else:
            self.image_file = None

        if len(param_list) >= 5:
            self.mask_file = param_list[4].reduce()
            if not isinstance(self.mask_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "alternative_sprites-block parameter 5 'mask_file' must be a string literal", self.mask_file.pos
                )
            if not self.bit_depth == 32:
                raise generic.ScriptError("A mask file may only be specified for 32 bpp sprites.", self.mask_file.pos)
        else:
            self.mask_file = None

        self.sprite_list = sprite_list

    def pre_process(self):
        if (self.bit_depth == 32 and not global_constants.allow_32bpp) or (
            self.zoom_level != 0 and not global_constants.allow_extra_zoom
        ):
            return
        block = sprite_container.SpriteContainer.resolve_sprite_block(self.name)
        block.add_sprite_data(
            self.sprite_list, self.image_file, self.pos, self.zoom_level, self.bit_depth, self.mask_file
        )

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Alternative sprites")
        generic.print_dbg(indentation + 2, "Replacement for sprite:", self.name)
        generic.print_dbg(indentation + 2, "Zoom level:", self.zoom_level)
        generic.print_dbg(indentation + 2, "Bit depth:", self.bit_depth)
        generic.print_dbg(indentation + 2, "Source:", self.image_file.value if self.image_file is not None else "None")
        generic.print_dbg(
            indentation + 2, "Mask source:", self.mask_file.value if self.mask_file is not None else "None"
        )

        generic.print_dbg(indentation + 2, "Sprites:")
        for sprite in self.sprite_list:
            sprite.debug_print(indentation + 4)

    def get_action_list(self):
        return []

    def __str__(self):
        params = [
            self.name,
            generic.reverse_lookup(global_constants.zoom_levels, self.zoom_level),
            generic.reverse_lookup(global_constants.bit_depths, self.bit_depth),
        ]
        if self.image_file is not None:
            params.append(self.image_file)
        if self.mask_file is not None:
            params.append(self.mask_file)
        ret = "alternative_sprites({}) {{\n".format(", ".join(str(p) for p in params))
        for sprite in self.sprite_list:
            ret += "\t{}\n".format(sprite)
        ret += "}\n"
        return ret



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml.actions import base_action, real_sprite, action2


class Action1(base_action.BaseAction):
    """
    Class representing an Action1

    @ivar feature: Feature of this action1
    @type feature: C{int}

    @ivar first_set: Number of the first sprite set in this action 1.
    @type first_set: C{int}

    @ivar num_sets: Number of (sprite) sets that follow this action 1.
    @type num_sets: C{int}

    @ivar num_ent: Number of sprites per set (e.g. (usually) 8 for vehicles)
    @type num_ent: C{int}
    """

    def __init__(self, feature, first_set, num_sets, num_ent):
        self.feature = feature
        self.first_set = first_set
        self.num_sets = num_sets
        self.num_ent = num_ent

    def write(self, file):
        if self.first_set == 0 and self.num_sets < 256:
            # <Sprite-number> * <Length> 01 <feature> <num-sets> <num-ent>
            file.start_sprite(6)
            file.print_bytex(1)
            file.print_bytex(self.feature)
            file.print_byte(self.num_sets)
            file.print_varx(self.num_ent, 3)
            file.newline()
            file.end_sprite()
        else:
            # <Sprite-number> * <Length> 01 <feature> 00 <first_set> <num-sets> <num-ent>
            file.start_sprite(12)
            file.print_bytex(1)
            file.print_bytex(self.feature)
            file.print_bytex(0)
            file.print_varx(self.first_set, 3)
            file.print_varx(self.num_sets, 3)
            file.print_varx(self.num_ent, 3)
            file.newline()
            file.end_sprite()


class SpritesetCollection(base_action.BaseAction):
    """
    A collection that contains multiple spritesets. All spritesets will be
    written to the same Action1, so they need to have the same number of sprites.

    @ivar feature: The feature number the action1 will get.
    @type feature: C{int}

    @ivar first_set: Number of the first sprite set in this action 1.
    @type first_set: C{int}

    @ivar num_sprites_per_spriteset: The number of sprites in each spriteset.
    @type num_sprites_per_spriteset: C{int}

    @ivar spritesets: A mapping from spritesets to indices. This allows for
                      quick lookup of whether a spriteset is already in this
                      collection. The indices are unique integers in the
                      range 0 .. len(spritesets) - 1.
    @type spritesets: C{dict} mapping L{SpriteSet} to C{int}.
    """

    def __init__(self, feature, first_set, num_sprites_per_spriteset):
        self.feature = feature
        self.first_set = first_set
        self.num_sprites_per_spriteset = num_sprites_per_spriteset
        self.spritesets = {}
        self.max_id = 0x3FFF if feature in action2.features_sprite_layout else 0xFFFF

    def skip_action7(self):
        return False

    def skip_action9(self):
        return False

    def skip_needed(self):
        return False

    def can_add(self, spriteset):
        """
        Test whether the given list of spritesets can be added to this collection.

        @param spriteset: The spriteset to test for addition.
        @type spriteset: L{SpriteSet}

        @return: True iff the given spriteset can be added to this collection.
        @rtype: C{bool}
        """
        assert self.first_set + 1 <= self.max_id
        if len(real_sprite.parse_sprite_data(spriteset)) != self.num_sprites_per_spriteset:
            return False
        return self.first_set + len(self.spritesets) + (1 if spriteset not in self.spritesets else 0) <= self.max_id

    def add(self, spriteset):
        """
        Add a spriteset to this collection.

        @param spriteset: The spriteset to add.
        @type spriteset: L{SpriteSet}

        @pre: can_add(spriteset).
        """
        assert self.can_add(spriteset)
        if spriteset not in self.spritesets:
            self.spritesets[spriteset] = len(self.spritesets)

    def get_index(self, spriteset):
        """
        Get the index of the given spriteset in the final action1.

        @param spriteset: The spriteset to get the index of.
        @type spriteset: L{SpriteSet}.

        @pre: The spriteset must have been previously added to this
              collection via #add.
        """
        assert spriteset in self.spritesets
        return self.first_set + self.spritesets[spriteset]

    def get_action_list(self):
        """
        Create a list of actions needed to write this collection to the output. This
        will generate a single Action1 and as many realsprites as needed.

        @return: A list of actions needed to represent this collection in a GRF.
        @rtype: C{list} of L{BaseAction}
        """
        actions = [Action1(self.feature, self.first_set, len(self.spritesets), self.num_sprites_per_spriteset)]
        for idx in range(len(self.spritesets)):
            for spriteset, spriteset_offset in self.spritesets.items():
                if idx == spriteset_offset:
                    actions.extend(real_sprite.parse_sprite_data(spriteset))
                    break
        return actions


"""
The list of collections per feature. add_to_action1 will try to reuse the
last collection as long as possible to reduce the duplication of sprites. As soon
as a spriteset with a different amount of sprites is added a new collection will
be created.
"""
spriteset_collections = {}


def add_to_action1(spritesets, feature, pos):
    """
    Add a list of spritesets to a spriteset collection. This will try to reuse
    one collection as long as possible and create a new one when needed.

    @param spritesets: List of spritesets that will be used by the next action2.
    @type spritesets: C{list} of L{SpriteSet}

    @param feature: Feature of the spritesets.
    @type feature: C{int}

    @param pos: Position reference to source.
    @type  pos: L{Position}

    @return: List of collections that needs to be added to the global action list.
    @rtype: C{list} of L{SpritesetCollection}.
    """
    if not spritesets:
        return []

    actions = []

    if feature not in spriteset_collections:
        spriteset_collections[feature] = [
            SpritesetCollection(feature, 0, len(real_sprite.parse_sprite_data(spritesets[0])))
        ]
        actions.append(spriteset_collections[feature][-1])

    current_collection = spriteset_collections[feature][-1]
    for spriteset in spritesets:
        for spriteset_collection in spriteset_collections[feature]:
            if spriteset in spriteset_collection.spritesets:
                continue
        if not current_collection.can_add(spriteset):
            spriteset_collections[feature].append(
                SpritesetCollection(
                    feature,
                    current_collection.first_set + len(current_collection.spritesets),
                    len(real_sprite.parse_sprite_data(spriteset)),
                )
            )
            current_collection = spriteset_collections[feature][-1]
            actions.append(current_collection)
        current_collection.add(spriteset)

    return actions


def get_action1_index(spriteset, feature):
    """
    Get the index of a spriteset in the action1. The given spriteset must have
    been added in the last call to #add_to_action1. Any new calls to
    #add_to_action1 may or may not allocate a new spriteset collection and as
    such make previous spritesets inaccessible.

    @param spriteset: The spriteset to get the index of.
    @type spriteset: L{SpriteSet}.

    @param feature: Feature of the spriteset.
    @type feature: C{int}

    @return: The index in the action1 of the given spriteset.
    @rtype: C{int}
    """
    assert feature in spriteset_collections
    for spriteset_collection in spriteset_collections[feature]:
        if spriteset in spriteset_collection.spritesets:
            return spriteset_collection.get_index(spriteset)
    assert False


def make_cb_failure_action1(feature):
    """
    Create an action1 that may be used for a callback failure
    If the last action1 is of the correct feature, no new action1 is needed
    Else, add a new action1 with 1 spriteset containing 0 sprites

    @param feature: Feature of the requested action 1
    @type feature: C{int}

    @return: List of actions to append (if any) and action1 index to use
    @rtype: C{tuple} of (C{list} of L{BaseAction}, C{int})
    """
    if feature in spriteset_collections:
        actions = []
    else:
        actions = [Action1(feature, 0, 1, 0)]
    return (actions, 0)  # Index is currently always 0, but will change with ext. A1



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic


class BaseStatement:
    """
    Base class for a statement (AST node) in NML.
    Note: All instance variables (except 'pos') are prefixed with "bs_" to avoid naming conflicts

    @ivar bs_name: Name of the statement type
    @type bs_name: C{str}

    @ivar pos: Position information
    @type pos: L{Position}

    @ivar bs_skipable: Whether the statement may be skipped using an if-block (default: True)
    @type bs_skipable: C{bool}

    @ivar bs_loopable: Whether the statement may be executed multiple times using a while-block (default: True)
    @type bs_loopable: C{bool}

    @pre: bs_skipable or not bs_loopable

    @ivar bs_in_item: Whether the statement may be part of an item block (default: False)
    @type bs_in_item: C{bool}

    @ivar bs_out_item: Whether the statement may appear outside of item blocks (default: True)
    @type bs_out_item: C{bool}
    """

    def __init__(self, name, pos, skipable=True, loopable=True, in_item=False, out_item=True):
        assert skipable or not loopable
        self.bs_name = name
        self.pos = pos
        self.bs_skipable = skipable
        self.bs_loopable = loopable
        self.bs_in_item = in_item
        self.bs_out_item = out_item

    def validate(self, scope_list):
        """
        Validate that this statement is in a valid location

        @param scope_list: List of nested blocks containing this statement
        @type scope_list: C{list} of L{BaseStatementList}
        """
        seen_item = False
        for scope in scope_list:
            if scope.list_type == BaseStatementList.LIST_TYPE_SKIP:
                if not self.bs_skipable:
                    raise generic.ScriptError(
                        "{} may not appear inside a conditional block.".format(self.bs_name), self.pos
                    )
            if scope.list_type == BaseStatementList.LIST_TYPE_LOOP:
                if not self.bs_loopable:
                    raise generic.ScriptError("{} may not appear inside a loop.".format(self.bs_name), self.pos)
            if scope.list_type == BaseStatementList.LIST_TYPE_ITEM:
                seen_item = True
                if not self.bs_in_item:
                    raise generic.ScriptError("{} may not appear inside an item block.".format(self.bs_name), self.pos)
        if not (seen_item or self.bs_out_item):
            raise generic.ScriptError("{} must appear inside an item block.".format(self.bs_name), self.pos)

    def register_names(self):
        """
        Called to register identifiers, that must be available before their definition.
        """
        pass

    def pre_process(self):
        """
        Called to do any pre-processing before the actual action generation.
        For example, to remove identifiers
        """
        pass

    def debug_print(self, indentation):
        """
        Print all AST information to the standard output

        @param indentation: Print all lines with at least C{indentation} spaces
        @type indentation: C{int}
        """
        raise NotImplementedError("debug_print must be implemented in BaseStatement-subclass {!r}".format(type(self)))

    def get_action_list(self):
        """
        Generate a list of NFO actions associated with this statement

        @return: A list of action
        @rtype: C{list} of L{BaseAction}
        """
        raise NotImplementedError(
            "get_action_list must be implemented in BaseStatement-subclass {!r}".format(type(self))
        )

    def __str__(self):
        """
        Generate a string representing this statement in valid NML-code.

        @return: An NML string representing this action
        @rtype: C{str}
        """
        raise NotImplementedError("__str__ must be implemented in BaseStatement-subclass {!r}".format(type(self)))


class BaseStatementList(BaseStatement):
    """
    Base class for anything that contains a list of statements

    @ivar list_type: Type of this list, used for validation logic
    @type list_type: C{int}, see constants below
    @pre list_type in (LIST_TYPE_NONE, LIST_TYPE_SKIP, LIST_TYPE_LOOP, LIST_TYPE_ITEM)

    @ivar statements: List of sub-statements in this block
    @type statements: C{list} of L{BaseStatement}
    """

    LIST_TYPE_NONE = 0
    LIST_TYPE_SKIP = 1
    LIST_TYPE_LOOP = 2
    LIST_TYPE_ITEM = 3

    def __init__(self, name, pos, list_type, statements, skipable=True, loopable=True, in_item=False, out_item=True):
        BaseStatement.__init__(self, name, pos, skipable, loopable, in_item, out_item)
        assert list_type in (self.LIST_TYPE_NONE, self.LIST_TYPE_SKIP, self.LIST_TYPE_LOOP, self.LIST_TYPE_ITEM)
        self.list_type = list_type
        self.statements = statements

    def validate(self, scope_list):
        new_list = scope_list + [self]
        for stmt in self.statements:
            stmt.validate(new_list)

    def register_names(self):
        for stmt in self.statements:
            stmt.register_names()

    def pre_process(self):
        for stmt in self.statements:
            stmt.pre_process()

    def debug_print(self, indentation):
        for stmt in self.statements:
            stmt.debug_print(indentation)

    def get_action_list(self):
        action_list = []
        for stmt in self.statements:
            action_list.extend(stmt.get_action_list())
        return action_list

    def __str__(self):
        res = ""
        for stmt in self.statements:
            res += "\t" + str(stmt).replace("\n", "\n\t")[0:-1]
        return res



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, grfstrings
from nml.actions import action6, actionD, base_action


class Action4(base_action.BaseAction):
    """
    Class representing a single action 4.
    Format: 04 <feature> <language-id> <num-ent> <offset> <text>

    @ivar feature: Feature of this action 4
    @type feature: C{int}

    @ivar lang: Language ID of the text (set as ##grflangid in the .lng file, 0x7F for default)
    @type lang: C{int}

    @ivar size: Size of the id, may be 1 (byte), 2 (word) or 3 (ext. byte)
    @type size: C{int}

    @ivar id: ID of the first string to write
    @type id: C{int}

    @ivar texts: List of strings to write
    @type texts: C{list} of C{str}
    """

    def __init__(self, feature, lang, size, id, texts):
        self.feature = feature
        self.lang = lang
        self.size = size
        self.id = id
        self.texts = texts

    def prepare_output(self, sprite_num):
        # To indicate a word value, bit 7 of the lang ID must be set
        if self.size == 2:
            self.lang = self.lang | 0x80

    def write(self, file):
        size = 4 + self.size
        for text in self.texts:
            size += grfstrings.get_string_size(text)
        file.start_sprite(size)
        file.print_bytex(4)
        file.print_bytex(self.feature)
        file.print_bytex(self.lang)
        file.print_bytex(len(self.texts))
        file.print_varx(self.id, self.size)
        for text in self.texts:
            file.print_string(text)
        file.newline()
        file.end_sprite()

    def skip_action9(self):
        return False


# List of various string ranges that may be used
# Attributes:
#  - random_id: If true, string IDs may be allocated randomly, else the ID has a special meaning and must be assigned
#               (e.g. for vehicles, string ID = vehicle ID)
#  - ids: List of free IDs, only needed if random_id is true. Whenever an ID is used, it's removed from the list
string_ranges = {
    0xC4: {"random_id": False},  # Station class names
    0xC5: {"random_id": False},  # Station names
    0xC9: {"random_id": False},  # House name
    # Misc. text ids, used for callbacks and such
    0xD0: {"random_id": True, "total": 0x400, "ids": list(range(0xD3FF, 0xCFFF, -1))},
    # Misc. persistent text ids, used to set properties.
    # Use Ids DC00..DCFF first to keep compatibility with older versions of OTTD.
    0xDC: {
        "random_id": True,
        "total": 0x2800,
        "ids": list(range(0xFFFF, 0xDFFF, -1)) + list(range(0xDBFF, 0xD7FF, -1)) + list(range(0xDFFF, 0xDBFF, -1)),
    },
}

# Mapping of string identifiers to D0xx/DCxx text ids
# This allows outputting strings only once, instead of everywhere they are used
used_strings = {
    0xD0: {},
    0xDC: {},
}


def print_stats():
    """
    Print statistics about used ids.
    """
    for t, l in string_ranges.items():
        if l["random_id"]:
            num_used = l["total"] - len(l["ids"])
            if num_used > 0:
                generic.print_info("{:02X}xx strings: {}/{}".format(t, num_used, l["total"]))


def get_global_string_actions():
    """
    Get a list of global string actions
    i.e. print all D0xx / DCxx texts at once

    @return: A list of all D0xx / DCxx action4s
    @rtype: C{list} of L{BaseAction}
    """
    texts = []
    actions = []
    for strings in used_strings.values():
        for feature_name, id in strings.items():
            feature, string_name = feature_name
            texts.append((0x7F, id, grfstrings.get_translation(string_name), feature))
            for lang_id in grfstrings.get_translations(string_name):
                texts.append((lang_id, id, grfstrings.get_translation(string_name, lang_id), feature))

    last_lang = -1
    last_id = -1
    last_feature = -1
    # Sort to have a deterministic ordering and to have as much consecutive IDs as possible
    texts.sort(key=lambda text: (-1 if text[0] == 0x7F else text[0], text[1]))

    for text in texts:
        str_lang, str_id, str_text, feature = text
        # If possible, append strings to the last action 4 instead of creating a new one
        if str_lang != last_lang or str_id - 1 != last_id or feature != last_feature or len(actions[-1].texts) == 0xFF:
            actions.append(Action4(feature, str_lang, 2, str_id, [str_text]))
        else:
            actions[-1].texts.append(str_text)
        last_lang = str_lang
        last_id = str_id
        last_feature = feature
    return actions


def get_string_action4s(feature, string_range, string, id=None):
    """
    Let a string from the lang files be used in the rest of NML.
    This may involve adding actions directly, but otherwise an ID is allocated and the string will be written later

    @param feature: Feature that uses the string
    @type feature: C{int}

    @param string_range: String range to use, either a value from L{string_ranges} or C{None} if N/A (item names)
    @type string_range: C{int} or C{None}

    @param string: String to parse
    @type string: L{expression.String}

    @param id: ID to use for this string, or C{None} if it will be allocated dynamically
               (random_id is true for the string range)
    @type id: L{Expression} or C{None}

    @return: A tuple of two values:
                - ID of the string (useful if allocated dynamically)
                - Resulting action list to be appended
    @rtype: C{tuple} of (C{int}, C{list} of L{BaseAction})
    """
    grfstrings.validate_string(string)
    write_action4s = True
    action6.free_parameters.save()
    actions = []

    mod = None
    if string_range is not None:
        size = 2
        if string_ranges[string_range]["random_id"]:
            # ID is allocated randomly, we will output the actions later
            write_action4s = False
            if (feature, string) in used_strings[string_range]:
                id_val = used_strings[string_range][(feature, string)]
            else:
                try:
                    id_val = string_ranges[string_range]["ids"].pop()
                    used_strings[string_range][(feature, string)] = id_val
                except IndexError:
                    raise generic.ScriptError(
                        "Unable to allocate ID for string, no more free IDs available (maximum is {:d})".format(
                            string_ranges[string_range]["total"]
                        ),
                        string.pos,
                    )
        else:
            # ID must be supplied
            assert id is not None
            assert isinstance(id, expression.ConstantNumeric)
            id_val = id.value | (string_range << 8)
    else:
        # Not a string range, so we must have an id
        assert id is not None
        size = 3 if feature <= 3 else 1
        if isinstance(id, expression.ConstantNumeric):
            id_val = id.value
        else:
            id_val = 0
            tmp_param, tmp_param_actions = actionD.get_tmp_parameter(id)
            actions.extend(tmp_param_actions)
            # Apply ID via action4 later
            mod = (tmp_param, 2 if feature <= 3 else 1, 5 if feature <= 3 else 4)

    if write_action4s:
        strings = [
            (lang_id, grfstrings.get_translation(string, lang_id)) for lang_id in grfstrings.get_translations(string)
        ]
        # Sort the strings for deterministic ordering and prepend the default language
        strings = [(0x7F, grfstrings.get_translation(string))] + sorted(strings, key=lambda lang_text: lang_text[0])

        for lang_id, text in strings:
            if mod is not None:
                act6 = action6.Action6()
                act6.modify_bytes(*mod)
                actions.append(act6)
            actions.append(Action4(feature, lang_id, size, id_val, [text]))

    action6.free_parameters.restore()

    return (id_val, actions)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic


class Assignment:
    """
    Simple storage container for a name / value pair.
    This class does not enforce any type information.
    Create a subclass to do this, or hard-code it in the parser.

    @ivar name: Name of the parameter / property / whatever to be set.
    @type name: Anything

    @ivar value: Value to assign to <name>
    @type value: Anything

    @ivar pos: Position information of the assignment
    @type pos: L{Position}
    """

    def __init__(self, name, value, pos):
        self.name = name
        self.value = value
        self.pos = pos

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Assignment")
        generic.print_dbg(indentation + 2, "Name:")
        self.name.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Value:")
        self.value.debug_print(indentation + 4)

    def __str__(self):
        return "{}: {};".format(self.name, self.value)


class UnitAssignment(Assignment):
    """
    Storage container where the value can have a unit.

    @ivar unit: Unit of the value, or not C{None}
    @type unit: L{Unit}
    """

    def __init__(self, name, value, unit, pos):
        Assignment.__init__(self, name, value, pos)
        self.unit = unit

    def debug_print(self, indentation):
        Assignment.debug_print(self, indentation)
        generic.print_dbg(indentation + 2, "Unit:")
        if self.unit is None:
            generic.print_dbg(indentation + 4, "None")
        else:
            generic.print_dbg(indentation + 4, self.unit)

    def __str__(self):
        if self.unit is None:
            return Assignment.__str__(self)
        else:
            return "{}: {} {};".format(self.name, self.value, self.unit.name)


class Range:
    """
    Storage container for a range of values (inclusive). This Contains
    a minimum value and optionally also a maximum value. If the maximum
    values is omitted, the minimum is also used as maximum.

    @ivar min: The minimum value of this range.
    @type min: L{Expression}

    @ivar max: The maximum value of this range.
    @type max: L{Expression} or C{None}
    """

    def __init__(self, min, max):
        self.min = min
        self.max = max

    def __str__(self):
        if self.max is None:
            return str(self.min)
        return "{} .. {}".format(self.min, self.max)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml.editors import extract_tables

output_file = "nml_kate.xml"

header_text = """\
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE language SYSTEM "language.dtd">
<!--
        This is a syntax highlighter for NML, the NewGRF markup language
        (http://dev.openttdcoop.org/nml) used to write NewGRF
        extensions for OpenTTD (http://www.openttd.org)

        The part on highlighting preprocessor instructions are taken from kate's c++ highlighter
        in the version 1.49 by Sebastian Pipping (webmaster@hartwork.org)
-->
<language name="NML" section="Sources" version="0.01" kateversion="2.4"
      extensions="*.nml;*.pnml;*.tnml"
      mimetype="text/x-nml"
      indenter="cstyle"
      author="Ingo von Borstel"
      license="GPL v2"
      priority="10">

  <highlighting>
    <list name="blockwords">
"""

feature_text = """\
    </list>
    <list name="features">
"""

builtin_text = """\
    </list>
    <list name="builtin">
"""

constant_text = """\
    </list>
    <list name="constants">
"""

tail_text = """\
    </list>

    <contexts>
      <!-- Normal text with the usual keywords and inserts -->
      <context name="Normal" attribute="Normal Text" lineEndContext="#stay">
        <DetectSpaces />
        <DetectChar context="AfterHash" char="#" firstNonSpace="true" lookAhead="true" />
        <keyword attribute="Block" context="#stay" String="blockwords" />
        <keyword attribute="Feature" context="#stay" String="features" />
        <keyword attribute="Built-in Function" context="#stay" String="builtin" />
        <keyword attribute="Constant" context="#stay" String="constants" />
        <DetectChar attribute="String" context="String" char="&quot;"/>
        <Detect2Chars attribute="Comment" context="Comment 1" char="/" char1="/" />
        <Detect2Chars attribute="Comment" context="Comment 2" char="/" char1="*" />
      </context>

      <!-- Strings within quotation marks -->
      <context attribute="String" lineEndContext="#pop" name="String">
        <LineContinue attribute="String" context="#stay"/>
        <HlCStringChar attribute="String Char" context="#stay"/>
        <DetectChar attribute="String" context="#pop" char="&quot;"/>
      </context>

      <!-- Strings starting with // -->
      <context attribute="Comment" lineEndContext="#pop" name="Comment 1">
        <LineContinue attribute="Comment" context="#stay"/>
        <DetectSpaces />
        <IncludeRules context="##Alerts" />
        <DetectIdentifier />
      </context>
      <!-- Multi-line strings like /* ... */ -->
      <context attribute="Comment" lineEndContext="#stay" name="Comment 2">
        <DetectSpaces />
        <Detect2Chars attribute="Comment" context="#pop" char="*" char1="/" endRegion="Comment"/>
        <IncludeRules context="##Alerts" />
        <DetectIdentifier />
      </context>

      <!-- Preprocessor commands starting with a hash - Main switch for preprocessor -->
      <context attribute="Error" lineEndContext="#pop" name="AfterHash">
        <!-- define,elif,else,endif,error,if,ifdef,ifndef,include,include_next,line,pragma,undef,warning -->
        <RegExpr attribute="Preprocessor" context="Preprocessor"
                 String="#\\s*if(?:def|ndef)?(?=\\s+\\S)" insensitive="true" beginRegion="PP" firstNonSpace="true" />
        <RegExpr attribute="Preprocessor" context="Preprocessor"
                 String="#\\s*endif" insensitive="true" endRegion="PP" firstNonSpace="true" />
        <RegExpr attribute="Preprocessor" context="Define"
                 String="#\\s*define.*((?=\\))" insensitive="true" firstNonSpace="true" />
        <RegExpr attribute="Preprocessor" context="Preprocessor"
                 String="#\\s*(?:el(?:se|if)|include(?:_next)?|define|undef|line|error|warning|pragma)"
                 insensitive="true" firstNonSpace="true" />
        <RegExpr attribute="Preprocessor" context="Preprocessor"
                 String="#\\s+[0-9]+" insensitive="true" firstNonSpace="true" />
      </context>
      <!-- Preprocessor instructions -->
      <context attribute="Preprocessor" lineEndContext="#pop" name="Preprocessor">
        <LineContinue attribute="Preprocessor" context="#stay"/>
        <RangeDetect attribute="Prep. Lib" context="#stay" char="&quot;" char1="&quot;"/>
        <RangeDetect attribute="Prep. Lib" context="#stay" char="&lt;" char1="&gt;"/>
        <IncludeRules context="##Doxygen" />
        <Detect2Chars attribute="Comment" context="Comment/Preprocessor" char="/" char1="*" beginRegion="Comment2" />
        <Detect2Chars attribute="Comment" context="Comment 1" char="/" char1="/"/>
      </context>
      <!-- Preprocessor #define -->
      <context attribute="Preprocessor" lineEndContext="#pop" name="Define">
        <LineContinue attribute="Preprocessor" context="#stay"/>
      </context>
      <!-- Preprocessor comments -->
      <context attribute="Comment" lineEndContext="#stay" name="Comment/Preprocessor">
        <DetectSpaces />
        <Detect2Chars attribute="Comment" context="#pop" char="*" char1="/" endRegion="Comment2" />
        <DetectIdentifier />
      </context>

    </contexts>

    <itemDatas>
      <itemData name="Normal Text"       defStyleNum="dsNormal" />
      <itemData name="Block"             defStyleNum="dsKeyword" />
      <itemData name="Feature"           defStyleNum="dsKeyword" color="#0095ff"
                                         selColor="#ffffff" bold="1" italic="0" spellChecking="false"/>
      <itemData name="Built-in Function" defStyleNum="dsDataType" spellChecking="false" />
      <itemData name="String"            defStyleNum="dsString"/>
      <itemData name="Preprocessor"      defStyleNum="dsOthers" spellChecking="false"/>
      <itemData name="Constant"          defStyleNum="dsOthers" color="#006400" italic="1" spellChecking="false" />
      <itemData name="Comment"           defStyleNum="dsComment"/>
    </itemDatas>
  </highlighting>

  <general>
    <comments>
      <comment name="singleLine" start="//" />
      <comment name="multiLine" start="/*" end="*/" region="Comment"/>
    </comments>
    <keywords casesensitive="1" />
  </general>

  </language>
<!--
// kate: space-indent on; indent-width 2; replace-tabs on;
-->
"""


def write_file(fname):
    with open(fname, "w") as file:
        file.write(header_text)
        for word in extract_tables.keywords:
            file.write("      <item> {} </item>\n".format(word))

        file.write(feature_text)
        for word in extract_tables.features:
            file.write("      <item> {} </item>\n".format(word))

        file.write(builtin_text)
        for word in extract_tables.functions:
            file.write("      <item> {} </item>\n".format(word))

        file.write(constant_text)
        for word in extract_tables.callback_names_table:
            file.write("      <item> {} </item>\n".format(word))

        file.write(tail_text)


def run():
    write_file("nml_kate.xml")



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

"""
Abstract base classes that implements common functionality for output classes
"""
import array
import io


class OutputBase:
    """
    Base class for output to a data file.

    The file is opened with L{open}. Once that is done, data can be written
    using the L{file} data member. When finished writing, the file should be
    closed with L{close}.

    Derived classes should implement L{open_file} to perform the actual opening
    of the file. L{assemble_file} is called to actually compose the output from
    possible multiple sources.

    @ivar filename: Name of the data file.
    @type filename: C{str}

    @ivar file: Memory output file handle, if opened.
    @type file: C{file} or C{None}
    """

    def __init__(self, filename):
        self.filename = filename
        self.file = None

    def open(self):
        """
        Open the output file. Data gets stored in-memory.
        """
        raise NotImplementedError("Implement me in {}".format(type(self)))

    def open_file(self):
        """
        Construct the file handle of the disk output file.

        @return: File handle of the opened file.
        @rtype: C{file}
        """
        raise NotImplementedError("Implement me in {}".format(type(self)))

    def assemble_file(self, real_file):
        """
        File is about to be closed, last chance to append data.

        @param real_file: Actual output stream.
        @type  real_file: C{io.IOBase}
        """
        real_file.write(self.file.getvalue())

    def discard(self):
        """
        Close the memory file, without writing to disk.
        """
        if isinstance(self.file, io.IOBase):
            self.file.close()
        self.file = None

    def close(self):
        """
        Close the memory file, copy collected output to the real file.
        """
        with self.open_file() as real_file:
            self.assemble_file(real_file)
        self.discard()

    def skip_sprite_checks(self):
        """
        Return whether sprites need detailed parsing.
        """
        return False

    def __enter__(self):
        """
        Allow `OutputBase` and subclasses to be used as context managers.
        """
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        """
        Allow `OutputBase` and subclasses to be used as context managers.
        """
        self.close()


class SpriteOutputBase(OutputBase):
    """
    Base class for output to a sprite file.

    @ivar in_sprite: Set to true if we are currently outputting a sprite.
                        Outputting anything when not in a sprite causes an assert.
    @type in_sprite: C{bool}

    @ivar byte_count: Number of bytes written in the current sprite.
    @type byte_count: C{int}

    @ivar expected_count: Number of bytes expected in the current sprite.
    @type expected_count: C{int}
    """

    def __init__(self, filename):
        OutputBase.__init__(self, filename)
        self.in_sprite = False
        self.expected_count = 0
        self.byte_count = 0

    def close(self):
        assert not self.in_sprite
        OutputBase.close(self)

    def prepare_byte(self, value):
        """
        Normalize the provided value to an unsigned byte value.

        @param value: Value to normalize.
        @type  value: C{int}

        @return: Normalized value (0..255).
        @rtype:  C{int}

        @precond: Must be outputting a sprite.
        """
        assert self.in_sprite
        if -0x80 <= value < 0:
            value += 0x100
        assert value >= 0 and value <= 0xFF
        self.byte_count += 1
        return value

    def prepare_word(self, value):
        """
        Normalize the provided value to an unsigned word value.

        @param value: Value to normalize.
        @type  value: C{int}

        @return: Normalized value (0..65535).
        @rtype:  C{int}

        @precond: Must be outputting a sprite.
        """
        assert self.in_sprite
        if -0x8000 <= value < 0:
            value += 0x10000
        assert value >= 0 and value <= 0xFFFF
        self.byte_count += 2
        return value

    def prepare_dword(self, value):
        """
        Normalize the provided value to an unsigned double word value.

        @param value: Value to normalize.
        @type  value: C{int}

        @return: Normalized value (0..0xFFFFFFFF).
        @rtype:  C{int}

        @precond: Must be outputting a sprite.
        """
        assert self.in_sprite
        if -0x80000000 <= value < 0:
            value += 0x100000000
        assert value >= 0 and value <= 0xFFFFFFFF
        self.byte_count += 4
        return value

    def print_varx(self, value, size):
        """
        Print a variable sized value.

        @param value: Value to output.
        @type  value: C{int}

        @param size: Size of the output (1..4), 3 means extended byte.
        @type  size: C{int}
        """
        if size == 1:
            self.print_bytex(value)
        elif size == 2:
            self.print_wordx(value)
        elif size == 3:
            self.print_bytex(0xFF)
            self.print_wordx(value)
        elif size == 4:
            self.print_dwordx(value)
        else:
            raise AssertionError()

    def print_bytex(self, byte, pretty_print=None):
        """
        Output an unsigned byte.

        @param byte: Value to output.
        @type  byte: C{int}
        """
        raise NotImplementedError("Implement print_bytex() in {}".format(type(self)))

    def print_wordx(self, byte):
        """
        Output an unsigned word (2 bytes).

        @param byte: Value to output.
        @type  byte: C{int}
        """
        raise NotImplementedError("Implement print_wordx() in {}".format(type(self)))

    def print_dwordx(self, byte):
        """
        Output an unsigned double word (4 bytes).

        @param byte: Value to output.
        @type  byte: C{int}
        """
        raise NotImplementedError("Implement print_dwordx() in {}".format(type(self)))

    def newline(self, msg="", prefix="\t"):
        """
        Output a line separator, prefixed with L{prefix}, C{"// "}, and the
        L{msg}, if the latter is not empty.

        @param msg: Optional message to output first.
        @type  msg: C{str}

        @param prefix: Additional white space in front of the comment.
        @type  prefix: C{str}
        """
        raise NotImplementedError("Implement newline() in {}".format(type(self)))

    def comment(self, msg):
        """
        Output a textual comment at a line by itself.

        @param msg: Comment message.
        @type  msg: C{str}

        @note: Only use if no bytes have been written to the current line.
        """
        raise NotImplementedError("Implement comment() in {}".format(type(self)))

    def start_sprite(self, expected_size, is_real_sprite=False):
        """
        Note to the output stream that a sprite is about to be written.

        @param expected_size: Expected size of the sprite data.
        @type  expected_size: C{int}

        @param is_real_sprite: Self-explanatory.
        @type  is_real_sprite: C{bool}
        """
        del is_real_sprite  # unused in base impl.
        assert not self.in_sprite
        self.in_sprite = True
        self.expected_count = expected_size
        self.byte_count = 0

    def end_sprite(self):
        """
        Note to the output stream that a sprite has been written. The number of
        bytes denoted as expected size with the L{start_sprite} call, should
        have been written.
        """
        assert self.in_sprite
        self.in_sprite = False
        self.newline()
        assert self.expected_count == self.byte_count, "Expected {:d} bytes to be written to sprite, got {:d}".format(
            self.expected_count, self.byte_count
        )


class TextOutputBase(OutputBase):
    """
    Base class for textual output.
    """

    def __init__(self, filename):
        OutputBase.__init__(self, filename)

    def open(self):
        self.file = io.StringIO()


class BinaryOutputBase(SpriteOutputBase):
    """
    Class for binary output.
    """

    def __init__(self, filename):
        SpriteOutputBase.__init__(self, filename)

    def open(self):
        self.file = array.array("B")

    def newline(self, msg="", prefix="\t"):
        pass

    def print_data(self, data):
        """
        Print a chunk of data in one go

        @param data: Data to output
        @type data: C{array}
        """
        self.byte_count += len(data)
        self.file.extend(data)

    def wb(self, byte):
        self.file.append(byte)

    def print_byte(self, value):
        value = self.prepare_byte(value)
        self.wb(value)

    def print_bytex(self, value, pretty_print=None):
        self.print_byte(value)

    def print_word(self, value):
        value = self.prepare_word(value)
        self.wb(value & 0xFF)
        self.wb(value >> 8)

    def print_wordx(self, value):
        self.print_word(value)

    def print_dword(self, value):
        value = self.prepare_dword(value)
        self.wb(value & 0xFF)
        self.wb((value >> 8) & 0xFF)
        self.wb((value >> 16) & 0xFF)
        self.wb(value >> 24)

    def print_dwordx(self, value):
        self.print_dword(value)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, global_constants

from .base_expression import Expression


class CargoExpression(Expression):
    def __init__(self, cargotype, value, pos):
        Expression.__init__(self, pos)
        assert isinstance(value, list)
        self.cargotype = cargotype
        self.value = value

    def cargolabel(self):
        for label, number in global_constants.cargo_numbers.items():
            if number == self.cargotype:
                return label
        raise AssertionError("Cargo expression with unregistered cargotype at " + str(self.pos))

    def debug_print(self, indentation):
        if self.value is None:
            generic.print_dbg(indentation, "{0} cargo {1}".format(self._debugname, self.cargolabel()))
        else:
            generic.print_dbg(indentation, "{0} cargo {1} with result:".format(self._debugname, self.cargolabel()))
            self.value.debug_print(indentation + 2)

    def __str__(self):
        return "{0}({1}, {2})".format(self._fnname, self.cargolabel(), str(self.value))

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        return self


class AcceptCargo(CargoExpression):
    _fnname = "accept_cargo"
    _debugname = "Accept"


class ProduceCargo(CargoExpression):
    _fnname = "produce_cargo"
    _debugname = "Produce"



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, nmlop
from nml.actions import action2, action2real, action2var, action6


class Action2Random(action2.Action2):
    def __init__(self, feature, name, pos, type_byte, count, triggers, randbit, nrand):
        action2.Action2.__init__(self, feature, name, pos)
        self.type_byte = type_byte
        self.count = count
        self.triggers = triggers
        self.randbit = randbit
        self.nrand = nrand
        self.choices = []

    def prepare_output(self, sprite_num):
        action2.Action2.prepare_output(self, sprite_num)
        for choice in self.choices:
            if isinstance(choice.result, expression.SpriteGroupRef):
                choice.result = choice.result.get_action2_id(self.feature)
            else:
                choice.result = choice.result.value | 0x8000

    def write(self, file):
        # <type> [<count>] <random-triggers> <randbit> <nrand> <set-ids>
        size = 4 + 2 * self.nrand + (self.count is not None)
        action2.Action2.write_sprite_start(self, file, size)
        file.print_bytex(self.type_byte)
        if self.count is not None:
            file.print_bytex(self.count)
        file.print_bytex(self.triggers)
        file.print_byte(self.randbit)
        file.print_bytex(self.nrand)
        file.newline()

        for choice in self.choices:
            for _ in range(0, choice.prob):
                file.print_wordx(choice.result)
            file.comment(choice.comment)
        file.end_sprite()


class RandomAction2Choice:
    def __init__(self, result, prob, comment):
        self.result = result
        self.prob = prob
        self.comment = comment


vehicle_random_types = {
    "SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 16, "triggers": True},
    "PARENT": {"type": 0x83, "param": 0, "first_bit": 0, "num_bits": 16, "triggers": False},
    "TILE": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 16, "triggers": False},
    "BACKWARD_SELF": {"type": 0x84, "param": 1, "first_bit": 0, "num_bits": 16, "triggers": False, "value": 0x00},
    "FORWARD_SELF": {"type": 0x84, "param": 1, "first_bit": 0, "num_bits": 16, "triggers": False, "value": 0x40},
    "BACKWARD_ENGINE": {"type": 0x84, "param": 1, "first_bit": 0, "num_bits": 16, "triggers": False, "value": 0x80},
    "BACKWARD_SAMEID": {"type": 0x84, "param": 1, "first_bit": 0, "num_bits": 16, "triggers": False, "value": 0xC0},
}
random_types = {
    0x00: vehicle_random_types,
    0x01: vehicle_random_types,
    0x02: vehicle_random_types,
    0x03: vehicle_random_types,
    0x04: {
        "SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 16, "triggers": True},
        "TILE": {"type": 0x80, "param": 0, "first_bit": 16, "num_bits": 4, "triggers": True},
    },
    0x05: {"SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 8, "triggers": False}},
    0x06: {},
    0x07: {"SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 8, "triggers": True}},
    0x08: {},
    0x09: {
        "SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 8, "triggers": True},
        "PARENT": {"type": 0x83, "param": 0, "first_bit": 0, "num_bits": 16, "triggers": True},
    },
    0x0A: {"SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 16, "triggers": False}},
    0x0B: {},
    0x0C: {},
    0x0D: {},
    0x0E: {},
    0x0F: {"SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 8, "triggers": False}},
    0x10: {"SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 2, "triggers": False}},
    0x11: {
        "SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 16, "triggers": False},
        "TILE": {"type": 0x80, "param": 0, "first_bit": 16, "num_bits": 4, "triggers": False},
    },
    0x12: {"SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 2, "triggers": False}},
    0x13: {"SELF": {"type": 0x80, "param": 0, "first_bit": 0, "num_bits": 2, "triggers": False}},
}


def parse_randomswitch_type(random_switch):
    """
    Parse the type of random switch to determine the type and random bits to use.

    @param random_switch: Random switch to parse the type of
    @type random_switch: L{RandomSwitch}

    @return: A tuple containing the following:
                - The type byte of the resulting random action2.
                - The value to use as <count>, None if N/A.
                - Expression to parse in a preceding switch-block, None if N/A.
                - The first random bit that should be used (often 0)
                - The number of random bits available
    @rtype: C{tuple} of (C{int}, C{int} or C{None}, L{Expression} or C{None}, C{int}, C{int})
    """
    # Extract some stuff we'll often need
    type_str = random_switch.type.value
    type_pos = random_switch.type.pos
    feature_val = next(iter(random_switch.feature_set))

    # Validate type name / param combination
    if type_str not in random_types[feature_val]:
        raise generic.ScriptError(
            "Invalid combination for random_switch feature {:d} and type '{}'. ".format(feature_val, type_str), type_pos
        )
    type_info = random_types[feature_val][type_str]

    count_expr = None
    if random_switch.type_count is None:
        # No param given
        if type_info["param"] == 1:
            raise generic.ScriptError(
                "Value '{}' for random_switch parameter 2 'type' requires a parameter.".format(type_str), type_pos
            )
        count = None
    else:
        # Param given
        if type_info["param"] == 0:
            raise generic.ScriptError(
                "Value '{}' for random_switch parameter 2 'type' should not have a parameter.".format(type_str),
                type_pos,
            )
        if (
            isinstance(random_switch.type_count, expression.ConstantNumeric)
            and 1 <= random_switch.type_count.value <= 15
        ):
            count = random_switch.type_count.value
        else:
            count = 0
            count_expr = nmlop.STO_TMP(random_switch.type_count, 0x100, type_pos)
        count = type_info["value"] | count

    if random_switch.triggers.value != 0 and not type_info["triggers"]:
        raise generic.ScriptError(
            "Triggers may not be set for random_switch feature {:d} and type '{}'. ".format(feature_val, type_str),
            type_pos,
        )

    # Determine type byte and random bits
    type_byte = type_info["type"]
    start_bit = type_info["first_bit"]
    bits_available = type_info["num_bits"]

    return type_byte, count, count_expr, start_bit, bits_available


def lookup_random_action2(sg_ref):
    """
    Lookup a sprite group reference to find the corresponding random action2

    @param sg_ref: Reference to random action2
    @type sg_ref: L{SpriteGroupRef}

    @return: Random action2 corresponding to this sprite group, or None if N/A
    @rtype: L{Action2Random} or C{None}
    """
    spritegroup = action2.resolve_spritegroup(sg_ref.name)
    act2 = spritegroup.random_act2
    assert act2 is None or isinstance(act2, Action2Random)
    return act2


def parse_randomswitch_dependencies(random_switch, start_bit, bits_available, nrand):
    """
    Handle the dependencies between random chains to determine the random bits to use

    @param random_switch: Random switch to parse
    @type random_switch: L{RandomSwitch}

    @param start_bit: First available random bit
    @type start_bit: C{int}

    @param bits_available: Number of random bits available
    @type bits_available: C{int}

    @param nrand: Number of random choices to use
    @type nrand: C{int}

    @return: A tuple of two values:
                - The first random bit to use
                - The number of random choices to use. This may be higher than the original amount passed as parameter
    @rtype: C{tuple} of (C{int}, C{int})
    """
    # Dependent random chains
    act2_to_copy = None
    for dep in random_switch.dependent:
        act2 = lookup_random_action2(dep)
        if act2 is None:
            continue  # May happen if said random switch is not used and therefore not parsed
        if act2_to_copy is not None:
            if act2_to_copy.randbit != act2.randbit:
                msg = (
                    "random_switch '{}' cannot be dependent on both '{}' and '{}'"
                    " as these are independent of each other."
                ).format(random_switch.name.value, act2_to_copy.name, act2.name)
                raise generic.ScriptError(msg, random_switch.pos)

            if act2_to_copy.nrand != act2.nrand:
                msg = (
                    "random_switch '{}' cannot be dependent on both '{}' and '{}'"
                    " as they don't use the same amount of random data."
                ).format(random_switch.name.value, act2_to_copy.name, act2.name)
                raise generic.ScriptError(msg, random_switch.pos)
        else:
            act2_to_copy = act2

    if act2_to_copy is not None:
        randbit = act2_to_copy.randbit
        if nrand > act2_to_copy.nrand:
            msg = "random_switch '{}' cannot be dependent on '{}' as it requires more random data."
            msg = msg.format(random_switch.name.value, act2_to_copy.name)
            raise generic.ScriptError(msg, random_switch.pos)

        nrand = act2_to_copy.nrand
    else:
        randbit = -1

    # Independent random chains
    possible_mask = ((1 << bits_available) - 1) << start_bit
    for indep in random_switch.independent:
        act2 = lookup_random_action2(indep)
        if act2 is None:
            continue  # May happen if said random switch is not used and therefore not parsed
        possible_mask &= ~((act2.nrand - 1) << act2.randbit)

    required_mask = nrand - 1
    if randbit != -1:
        # randbit has already been determined. Check that it is suitable
        if possible_mask & (required_mask << randbit) != (required_mask << randbit):
            msg = (
                "Combination of dependence on and independence from"
                " random_switches is not possible for random_switch '{}'."
            ).format(random_switch.name.value)
            raise generic.ScriptError(msg, random_switch.pos)
    else:
        # find a suitable randbit
        for i in range(start_bit, bits_available + start_bit):
            if possible_mask & (required_mask << i) == (required_mask << i):
                randbit = i
                break
        else:
            msg = "Independence of all given random_switches is not possible for random_switch '{}'."
            msg = msg.format(random_switch.name.value)
            raise generic.ScriptError(msg, random_switch.pos)

    return randbit, nrand


def parse_randomswitch(random_switch):
    """
    Parse a randomswitch block into actions

    @param random_switch: RandomSwitch block to parse
    @type random_switch: L{RandomSwitch}

    @return: List of actions
    @rtype: C{list} of L{BaseAction}
    """
    action_list = action2real.create_spriteset_actions(random_switch)
    feature = next(iter(random_switch.feature_set))
    type_byte, count, count_expr, start_bit, bits_available = parse_randomswitch_type(random_switch)

    total_prob = sum(choice.probability.value for choice in random_switch.choices)
    assert total_prob > 0
    nrand = 1
    while nrand < total_prob:
        nrand <<= 1

    # Verify that enough random data is available
    if min(1 << bits_available, 0x80) < nrand:
        msg = "The maximum sum of all random_switch probabilities is {:d}, encountered {:d}."
        msg = msg.format(min(1 << bits_available, 0x80), total_prob)
        raise generic.ScriptError(msg, random_switch.pos)

    randbit, nrand = parse_randomswitch_dependencies(random_switch, start_bit, bits_available, nrand)

    random_action2 = Action2Random(
        feature,
        random_switch.name.value,
        random_switch.pos,
        type_byte,
        count,
        random_switch.triggers.value,
        randbit,
        nrand,
    )
    random_switch.random_act2 = random_action2

    action6.free_parameters.save()
    act6 = action6.Action6()
    offset = 8 if count is not None else 7

    # divide the 'extra' probabilities in an even manner
    i = 0
    resulting_prob = {c: c.probability.value for c in random_switch.choices}
    while i < (nrand - total_prob):
        best_choice = None
        best_ratio = 0
        for choice in random_switch.choices:
            # float division, so 9 / 10 = 0.9
            ratio = choice.probability.value / float(resulting_prob[choice] + 1)
            if ratio > best_ratio:
                best_ratio = ratio
                best_choice = choice
        assert best_choice is not None
        resulting_prob[best_choice] += 1
        i += 1

    for choice in random_switch.choices:
        res_prob = resulting_prob[choice]
        result, comment = action2var.parse_result(
            choice.result.value,
            action_list,
            act6,
            offset,
            random_action2,
            None,
            0x8A if type_byte == 0x83 else 0x89,
            res_prob,
        )
        offset += res_prob * 2
        comment = "({:d}/{:d}) -> ({:d}/{:d}): ".format(choice.probability.value, total_prob, res_prob, nrand) + comment
        random_action2.choices.append(RandomAction2Choice(result, res_prob, comment))

    if len(act6.modifications) > 0:
        action_list.append(act6)

    action_list.append(random_action2)
    if count_expr is None:
        random_switch.set_action2(random_action2, feature)
    else:
        # Create intermediate varaction2 to compute parameter for type 0x84
        varaction2 = action2var.Action2Var(
            feature, "{}@registers".format(random_switch.name.value), random_switch.pos, 0x89
        )
        varact2parser = action2var.Varaction2Parser(feature)
        varact2parser.parse_expr(count_expr)
        varaction2.var_list = varact2parser.var_list
        action_list.extend(varact2parser.extra_actions)
        extra_act6 = action6.Action6()
        for mod in varact2parser.mods:
            extra_act6.modify_bytes(mod.param, mod.size, mod.offset + 4)
        if len(extra_act6.modifications) > 0:
            action_list.append(extra_act6)
        ref = expression.SpriteGroupRef(random_switch.name, [], None, random_action2)
        varaction2.ranges.append(
            action2var.VarAction2Range(expression.ConstantNumeric(0), expression.ConstantNumeric(0), ref, "")
        )
        varaction2.default_result = ref
        varaction2.default_comment = ""

        # Add two references (default + range)
        action2.add_ref(ref, varaction2)
        action2.add_ref(ref, varaction2)
        random_switch.set_action2(varaction2, feature)
        action_list.append(varaction2)

    action6.free_parameters.restore()
    return action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from itertools import chain

from nml import generic
from nml.actions import action2

from .base_expression import Expression, Type


class SpriteGroupRef(Expression):
    """
    Container for a reference to a sprite group / layout

    @ivar name: Name of the referenced item
    @type name: L{Identifier}

    @ivar param_list: List of parameters to be passed
    @type param_list: C{list} of L{Expression}

    @ivar pos: Position of this reference
    @type pos: L{Position}

    @ivar act2: Action2 that is the target of this reference
                    To be used for action2s that have no direct equivalent in the AST
    @type act2: L{Action2}
    """

    def __init__(self, name, param_list, pos, act2=None):
        super().__init__(pos)
        self.name = name
        self.param_list = param_list
        self.act2 = act2
        self.is_procedure = False

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Reference to:", self.name)
        if len(self.param_list) != 0:
            generic.print_dbg(indentation, "Parameters:")
            for p in self.param_list:
                p.debug_print(indentation + 2)

    def __str__(self):
        if self.param_list:
            return "{}({})".format(self.name, ", ".join(str(x) for x in self.param_list))
        return str(self.name)

    def get_action2_id(self, feature):
        """
        Get the action2 set-ID that this reference maps to

        @param feature: Feature of the action2
        @type feature: C{int}

        @return: The set ID
        @rtype: C{int}
        """
        if self.act2 is not None:
            return self.act2.id
        if self.name.value == "CB_FAILED":
            return 0  # 0 serves as a failed CB result because it is never used
        try:
            spritegroup = action2.resolve_spritegroup(self.name)
        except generic.ScriptError:
            raise AssertionError("Illegal action2 reference '{}' encountered.".format(self.name.value))

        return spritegroup.get_action2(feature).id

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        if self.name.value != "CB_FAILED" and not self.is_procedure:
            spritegroup = action2.resolve_spritegroup(self.name)
            if spritegroup.optimise():
                return spritegroup.optimised
        return self

    def supported_by_action2(self, raise_error):
        return True

    def collect_references(self):
        return list(chain([self], *(p.collect_references() for p in self.param_list)))

    def is_read_only(self):
        if self.name.value != "CB_FAILED":
            spritegroup = action2.resolve_spritegroup(self.name)
            return spritegroup.is_read_only()
        return True

    def type(self):
        if self.is_procedure:
            return Type.INTEGER
        return Type.SPRITEGROUP_REF

    def __eq__(self, other):
        return (
            other is not None
            and isinstance(other, SpriteGroupRef)
            and other.name == self.name
            and other.param_list == self.param_list
        )

    def __hash__(self):
        return hash(self.name) ^ hash(tuple(self.param_list))



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, global_constants

from .base_expression import ConstantNumeric, Expression
from .string_literal import StringLiteral

ignore_all_invalid_ids = False


def default_id_func(name, x, pos):
    """
    Default id conversion function.

    @param name: Key in the containing dictionary.
    @type  name: Any, unused (to match other id_dicts callbacks)

    @param x: Value to convert.
    @type  x: C{str}, C{int}, or C{float}

    @param pos: Position of the id.
    @type  pos: L{Position}

    @return: Expression of the id.
    @rtype:  L{Expression}
    """
    del name  # unused
    if isinstance(x, str):
        return StringLiteral(x, pos)
    else:
        return ConstantNumeric(x, pos)


class Identifier(Expression):
    def __init__(self, value, pos=None):
        Expression.__init__(self, pos)
        self.value = value
        if value in global_constants.identifier_refcount:
            global_constants.identifier_refcount[value] += 1
        else:
            global_constants.identifier_refcount[value] = 0

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "ID:", self.value)

    def __str__(self):
        return self.value

    def reduce(self, id_dicts=None, unknown_id_fatal=True, search_func_ptr=False):
        for id_dict in id_dicts or []:
            if isinstance(id_dict, tuple):
                id_d, func = id_dict
            else:
                id_d, func = id_dict, default_id_func

            if self.value in id_d:
                if search_func_ptr:
                    # Do not reduce function pointers, since they have no (numerical) value
                    return func(self.value, id_d[self.value], self.pos)
                else:
                    return func(self.value, id_d[self.value], self.pos).reduce(id_dicts)

        if unknown_id_fatal and not ignore_all_invalid_ids:
            raise generic.ScriptError("Unrecognized identifier '" + self.value + "' encountered", self.pos)

        return self

    def supported_by_actionD(self, raise_error):
        if raise_error:
            raise generic.ScriptError("Unknown identifier '{}'".format(self.value), self.pos)

        return False

    def __eq__(self, other):
        return other is not None and isinstance(other, Identifier) and self.value == other.value

    def __hash__(self):
        return hash(self.value)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import nmlop
from nml.actions import action6, actionD, base_action, real_sprite


class ActionA(base_action.BaseAction):
    """
    Action class for Action A (sprite replacement)

    @ivar sets: List of sprite collections to be replaced.
    @type sets: C{list} of (C{int}, C{int})-tuples
    """

    def __init__(self, sets):
        self.sets = sets

    def write(self, file):
        # <Sprite-number> * <Length> 0A <num-sets> [<num-sprites> <first-sprite>]+
        size = 2 + 3 * len(self.sets)
        file.start_sprite(size)
        file.print_bytex(0x0A)
        file.print_byte(len(self.sets))
        for num, first in self.sets:
            file.print_byte(num)
            file.print_word(first)
        file.newline()
        file.end_sprite()


def parse_actionA(replaces):
    """
    Parse replace-block to ActionA.

    @param replaces: Replace-block to parse.
    @type  replaces: L{ReplaceSprite}
    """
    action_list = []
    action6.free_parameters.save()
    act6 = action6.Action6()

    real_sprite_list = real_sprite.parse_sprite_data(replaces)
    block_list = []
    total_sprites = len(real_sprite_list)
    offset = 2  # Skip 0A and <num-sets>
    sprite_offset = 0  # Number of sprites already covered by previous [<num-sprites> <first-sprite>]-pairs

    while total_sprites > 0:
        this_block = min(total_sprites, 255)  # number of sprites in this block
        total_sprites -= this_block
        offset += 1  # Skip <num-sprites>

        first_sprite = replaces.start_id  # number of first sprite
        if sprite_offset != 0:
            first_sprite = nmlop.ADD(first_sprite, sprite_offset).reduce()
        first_sprite, offset = actionD.write_action_value(first_sprite, action_list, act6, offset, 2)
        block_list.append((this_block, first_sprite.value))

        sprite_offset += this_block  # increase first-sprite for next block

    if len(act6.modifications) > 0:
        action_list.append(act6)
    action6.free_parameters.restore()

    action_list.append(ActionA(block_list))
    action_list.extend(real_sprite_list)

    return action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic
from nml.actions import action1, action2
from nml.ast import general


class Action2Real(action2.Action2):
    def __init__(self, feature, name, pos, loaded_list, loading_list):
        action2.Action2.__init__(self, feature, name, pos)
        self.loaded_list = loaded_list
        self.loading_list = loading_list

    def write(self, file):
        size = 2 + 2 * len(self.loaded_list) + 2 * len(self.loading_list)
        action2.Action2.write_sprite_start(self, file, size)
        file.print_byte(len(self.loaded_list))
        file.print_byte(len(self.loading_list))
        file.newline()
        for i in self.loaded_list:
            file.print_word(i)
        file.newline()
        for i in self.loading_list:
            file.print_word(i)
        file.newline()
        file.end_sprite()


def get_real_action2s(spritegroup, feature):
    loaded_list = []
    loading_list = []
    actions = []

    if feature not in action2.features_sprite_group:
        raise generic.ScriptError(
            "Sprite groups that combine sprite sets are not supported for feature '{}'.".format(
                general.feature_name(feature)
            ),
            spritegroup.pos,
        )

    # First make sure that all referenced real sprites are put in a single action1
    spriteset_list = []
    for view in spritegroup.spriteview_list:
        spriteset_list.extend([action2.resolve_spritegroup(sg_ref.name) for sg_ref in view.spriteset_list])
        if feature == 0x04:
            if view.name.value not in ["little", "lots"]:
                raise generic.ScriptError("Unexpected '{}' (list of) sprite set(s).".format(view.name), view.pos)
            view.name.value = "loading" if view.name.value == "lots" else "loaded"
    actions.extend(action1.add_to_action1(spriteset_list, feature, spritegroup.pos))

    view_names = sorted(view.name.value for view in spritegroup.spriteview_list)
    if feature in (0x00, 0x01, 0x02, 0x03):
        if view_names != sorted(["loading", "loaded"]):
            raise generic.ScriptError("Expected a 'loading' and a 'loaded' (list of) sprite set(s).", spritegroup.pos)
    elif feature == 0x04:
        if "loading" not in view_names:
            raise generic.ScriptError("Expected at least a 'lots' (list of) sprite set(s).", spritegroup.pos)
    elif feature in (0x05, 0x0B, 0x0D, 0x10):
        msg = (
            "Sprite groups for feature {:02X} will not be supported in the future, as they are no longer needed."
            " Directly refer to sprite sets instead."
        ).format(feature)
        generic.print_warning(generic.Warning.GENERIC, msg, spritegroup.pos)
        if view_names != ["default"]:
            raise generic.ScriptError("Expected only a 'default' (list of) sprite set(s).", spritegroup.pos)

    for view in spritegroup.spriteview_list:
        if len(view.spriteset_list) == 0:
            raise generic.ScriptError("Expected at least one sprite set, encountered 0.", view.pos)
        for set_ref in view.spriteset_list:
            spriteset = action2.resolve_spritegroup(set_ref.name)
            action1_index = action1.get_action1_index(spriteset, feature)
            if view.name.value == "loading":
                loading_list.append(action1_index)
            else:
                loaded_list.append(action1_index)

    if len(loaded_list) >= 0x80:
        raise generic.ScriptError("Too many 'loaded' sprite set(s).", spritegroup.pos)
    if len(loading_list) >= 0xFF:
        raise generic.ScriptError("Too many 'loading' sprite set(s).", spritegroup.pos)

    actions.append(
        Action2Real(
            feature,
            spritegroup.name.value + " - feature {:02X}".format(feature),
            spritegroup.pos,
            loaded_list,
            loading_list,
        )
    )
    spritegroup.set_action2(actions[-1], feature)
    return actions


def make_simple_real_action2(feature, name, pos, action1_index):
    """
    Make a simple real action2, referring to only 1 action1 sprite set

    @param feature: Feature of the needed action2
    @type feature: C{int}

    @param name: Name of the action2
    @type name: C{str}

    @param pos: Positional context.
    @type  pos: L{Position}

    @param action1_index: Index of the action1 sprite set to use
    @type action1_index: C{int}

    @return: The created real action2
    @rtype: L{Action2Real}
    """
    return Action2Real(
        feature, name, pos, [action1_index] if feature != 0x04 else [], [action1_index] if feature <= 0x04 else []
    )


def create_spriteset_actions(spritegroup):
    """
    Create action2s for directly-referenced sprite sets

    @param spritegroup: Spritegroup to create the sprite sets for
    @type spritegroup: L{ASTSpriteGroup}

    @return: Resulting list of actions
    @rtype: C{list} of L{BaseAction}
    """
    action_list = []
    # Iterate over features first for more efficient action1s
    for feature in spritegroup.feature_set:
        if len(spritegroup.used_sprite_sets) != 0 and feature not in action2.features_sprite_group:
            raise generic.ScriptError(
                "Directly referring to sprite sets is not possible for feature {:02X}".format(feature), spritegroup.pos
            )
        for spriteset in spritegroup.used_sprite_sets:
            if spriteset.has_action2(feature):
                continue
            action_list.extend(action1.add_to_action1([spriteset], feature, spritegroup.pos))

            real_action2 = make_simple_real_action2(
                feature,
                spriteset.name.value + " - feature {:02X}".format(feature),
                spritegroup.pos,
                action1.get_action1_index(spriteset, feature),
            )
            action_list.append(real_action2)
            spriteset.set_action2(real_action2, feature)
    return action_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action5, actionA
from nml.ast import base_statement, sprite_container


class ReplaceSprite(base_statement.BaseStatement, sprite_container.SpriteContainer):
    """
    AST node for a 'replace' block.
    NML syntax: replace [name] (start_id[, default_file]) { ..real sprites.. }

    @ivar start_id: First sprite to replace. Extracted from C{param_list} during pre-processing.
    @type start_id: C{Expression}

    @ivar image_file: Default image file to use for sprites.
    @type image_file: C{None} if not specified, else L{StringLiteral}

    @ivar sprite_list: List of real sprites to use
    @type sprite_list: Heterogeneous C{list} of L{RealSprite}, L{TemplateUsage}
    """

    def __init__(self, param_list, sprite_list, name, pos):
        base_statement.BaseStatement.__init__(self, "replace-block", pos)
        sprite_container.SpriteContainer.__init__(self, "replace-block", name)

        num_params = len(param_list)
        if not (1 <= num_params <= 2):
            raise generic.ScriptError("replace-block requires 1 or 2 parameters, encountered " + str(num_params), pos)
        self.start_id = param_list[0]
        if num_params >= 2:
            self.image_file = param_list[1].reduce()
            if not isinstance(self.image_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "replace-block parameter 2 'file' must be a string literal", self.image_file.pos
                )
        else:
            self.image_file = None
        self.sprite_list = sprite_list
        self.add_sprite_data(self.sprite_list, self.image_file, pos)

    def pre_process(self):
        self.start_id = self.start_id.reduce(global_constants.const_list)
        if isinstance(self.start_id, expression.ConstantNumeric):
            generic.check_range(
                self.start_id.value,
                0,
                4895 - len(self.sprite_list),
                "replace-block parameter 1 'start'",
                self.start_id.pos,
            )

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Replace sprites starting at")
        self.start_id.debug_print(indentation + 2)

        generic.print_dbg(indentation + 2, "Source:", self.image_file.value if self.image_file is not None else "None")
        if self.block_name is not None:
            generic.print_dbg(indentation + 2, "Name:", self.block_name.value)

        generic.print_dbg(indentation + 2, "Sprites:")
        for sprite in self.sprite_list:
            sprite.debug_print(indentation + 4)

    def get_action_list(self):
        return actionA.parse_actionA(self)

    def __str__(self):
        name = str(self.block_name) if self.block_name is not None else ""
        def_file = "" if self.image_file is None else ", " + str(self.image_file)
        ret = "replace {}({}{}) {{\n".format(name, self.start_id, def_file)
        for sprite in self.sprite_list:
            ret += "\t{}\n".format(sprite)
        ret += "}\n"
        return ret


class ReplaceNewSprite(base_statement.BaseStatement, sprite_container.SpriteContainer):
    """
    AST node for a 'replacenew' block.
    NML syntax: replacenew [name](type[, default_file[, offset]]) { ..real sprites.. }

    @ivar type: Type of sprites to replace.
    @type type: L{Identifier}

    @ivar image_file: Default image file to use for sprites.
    @type image_file: C{None} if not specified, else L{StringLiteral}

    @ivar offset: Offset into the block of sprites.
    @type offset: C{int}

    @ivar sprite_list: List of real sprites to use
    @type sprite_list: Heterogeneous C{list} of L{RealSprite}, L{TemplateUsage}
    """

    def __init__(self, param_list, sprite_list, name, pos):
        base_statement.BaseStatement.__init__(self, "replacenew-block", pos)
        sprite_container.SpriteContainer.__init__(self, "replacenew-block", name)
        num_params = len(param_list)
        if not (1 <= num_params <= 3):
            raise generic.ScriptError(
                "replacenew-block requires 1 to 3 parameters, encountered " + str(num_params), pos
            )

        self.type = param_list[0]
        if not isinstance(self.type, expression.Identifier):
            raise generic.ScriptError(
                "replacenew parameter 'type' must be an identifier of a sprite replacement type", self.type.pos
            )

        if num_params >= 2:
            self.image_file = param_list[1].reduce()
            if not isinstance(self.image_file, expression.StringLiteral):
                raise generic.ScriptError(
                    "replacenew-block parameter 2 'file' must be a string literal", self.image_file.pos
                )
        else:
            self.image_file = None

        if num_params >= 3:
            self.offset = param_list[2].reduce_constant().value
            generic.check_range(self.offset, 0, 0xFFFF, "replacenew-block parameter 3 'offset'", param_list[2].pos)
        else:
            self.offset = 0

        self.sprite_list = sprite_list
        self.add_sprite_data(self.sprite_list, self.image_file, pos)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Replace sprites for new features of type", self.type)
        generic.print_dbg(indentation + 2, "Offset:  ", self.offset)
        generic.print_dbg(
            indentation + 2, "Source:  ", self.image_file.value if self.image_file is not None else "None"
        )
        if self.block_name is not None:
            generic.print_dbg(indentation + 2, "Name:", self.block_name.value)

        generic.print_dbg(indentation + 2, "Sprites:")
        for sprite in self.sprite_list:
            sprite.debug_print(indentation + 4)

    def get_action_list(self):
        return action5.parse_action5(self)

    def __str__(self):
        name = str(self.block_name) if self.block_name is not None else ""
        params = [self.type]
        if self.image_file is not None:
            params.append(self.image_file)
            if self.offset != 0:
                params.append(self.offset)
        ret = "replacenew {}({}) {{\n".format(name, ", ".join(str(param) for param in params))
        for sprite in self.sprite_list:
            ret += "\t{}\n".format(sprite)
        ret += "}\n"
        return ret



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import codecs
import optparse
import os
import sys

from nml import (
    generic,
    global_constants,
    grfstrings,
    output_dep,
    output_grf,
    output_nfo,
    output_nml,
    palette,
    parser,
    spritecache,
    spriteencoder,
    version_info,
)
from nml.actions import (
    action0,
    action1,
    action2,
    action2layout,
    action2var,
    action4,
    action6,
    action7,
    action8,
    action11,
    actionF,
    real_sprite,
    sprite_count,
)
from nml.ast import grf

try:
    from PIL import Image
except ImportError:
    # Image is required only when using graphics
    Image = None
    pass

developmode = False  # Give 'nice' error message instead of a stack dump.

version = version_info.get_nml_version()


def parse_cli(argv):
    """
    Parse the command line, and process options.

    @return: Options, and input filename (if not supplied, use C{sys.stdin}).
    @rtype:  C{tuple} (C{Object}, C{str} or C{None})
    """
    usage = "Usage: %prog [options] <filename>\n" "Where <filename> is the nml file to parse"

    opt_parser = optparse.OptionParser(usage=usage, version=version_info.get_cli_version())
    opt_parser.set_defaults(
        debug=False,
        crop=False,
        compress=True,
        outputs=[],
        start_sprite_num=0,
        custom_tags="custom_tags.txt",
        lang_dir="lang",
        default_lang="english.lng",
        cache_dir=".nmlcache",
        forced_palette="ANY",
        quiet=False,
        md5_filename=None,
        keep_orphaned=True,
        verbosity=generic.verbosity_level,
        debug_parser=False,
        allow_extra_zoom=True,
        allow_32bpp=True,
        disable_palette_validation=False,
        list_unused_strings=False,
    )
    opt_parser.add_option("-d", "--debug", action="store_true", dest="debug", help="write the AST to stdout")
    opt_parser.add_option("-s", "--stack", action="store_true", dest="stack", help="Dump stack when an error occurs")
    opt_parser.add_option("--grf", dest="grf_filename", metavar="<file>", help="write the resulting grf to <file>")
    opt_parser.add_option(
        "--md5", dest="md5_filename", metavar="<file>", help="Write an md5sum of the resulting grf to <file>"
    )
    opt_parser.add_option("--nfo", dest="nfo_filename", metavar="<file>", help="write nfo output to <file>")
    opt_parser.add_option(
        "-M",
        action="store_true",
        dest="dep_check",
        help=(
            "output a rule suitable for make describing"
            " the graphics dependencies of the main grf file (requires input file or --grf)"
        ),
    )
    opt_parser.add_option(
        "--MF",
        dest="dep_filename",
        metavar="<file>",
        help="When used with -M, specifies a file to write the dependencies to",
    )
    opt_parser.add_option(
        "--MT",
        dest="depgrf_filename",
        metavar="<file>",
        help="target of the rule emitted by dependency generation (requires -M)",
    )
    opt_parser.add_option(
        "-c", action="store_true", dest="crop", help="crop extraneous transparent blue from real sprites"
    )
    opt_parser.add_option("-u", action="store_false", dest="compress", help="save uncompressed data in the grf file")
    opt_parser.add_option("--nml", dest="nml_filename", metavar="<file>", help="write optimized nml to <file>")
    opt_parser.add_option(
        "-o", "--output", dest="outputs", action="append", metavar="<file>", help="write output(nfo/grf) to <file>"
    )
    opt_parser.add_option(
        "-t",
        "--custom-tags",
        dest="custom_tags",
        metavar="<file>",
        help="Load custom tags from <file> [default: %default]",
    )
    opt_parser.add_option(
        "-l",
        "--lang-dir",
        dest="lang_dir",
        metavar="<dir>",
        help="Load language files from directory <dir> [default: %default]",
    )
    opt_parser.add_option(
        "--default-lang",
        dest="default_lang",
        metavar="<file>",
        help="The default language is stored in <file> [default: %default]",
    )
    opt_parser.add_option(
        "--start-sprite",
        action="store",
        type="int",
        dest="start_sprite_num",
        metavar="<num>",
        help=(
            "Set the first sprite number to write"
            " (do not use except when you output nfo that you want to include in other files)"
        ),
    )
    opt_parser.add_option(
        "-p",
        "--palette",
        dest="forced_palette",
        metavar="<palette>",
        choices=["DEFAULT", "LEGACY", "DOS", "WIN", "ANY"],
        help="Force nml to use the palette <pal> [default: %default]. Valid values are 'DEFAULT', 'LEGACY', 'ANY'",
    )
    opt_parser.add_option(
        "--quiet", action="store_true", dest="quiet", help="Disable all warnings. Errors will be printed normally."
    )
    opt_parser.add_option(
        "-n",
        "--no-cache",
        action="store_true",
        dest="no_cache",
        help="Disable caching of sprites in .cache[index] files, which may reduce compilation time.",
    )
    opt_parser.add_option(
        "--cache-dir",
        dest="cache_dir",
        metavar="<dir>",
        help="Cache files are stored in directory <dir> [default: %default]",
    )
    opt_parser.add_option(
        "--clear-orphaned",
        action="store_false",
        dest="keep_orphaned",
        help="Remove unused/orphaned items from cache files.",
    )
    opt_parser.add_option(
        "--verbosity",
        type="int",
        dest="verbosity",
        metavar="<level>",
        help="Set the verbosity level for informational output. [default: %default, max: {}]".format(
            generic.VERBOSITY_MAX
        ),
    )
    opt_parser.add_option(
        "-D", "--debug-parser", action="store_true", dest="debug_parser", help="Enable debug mode for parser."
    )
    opt_parser.add_option(
        "--no-optimisation-warning",
        action="append_const",
        dest="disable_warning",
        const=generic.Warning.OPTIMISATION,
        help="Disable optimisation warnings",
    )
    opt_parser.add_option(
        "--no-deprecation-warning",
        action="append_const",
        dest="disable_warning",
        const=generic.Warning.DEPRECATION,
        help="Disable deprecation warnings",
    )
    opt_parser.add_option(
        "--no-extra-zoom", action="store_false", dest="allow_extra_zoom", help="Skip extra zoom alternative sprites"
    )
    opt_parser.add_option("--no-32bpp", action="store_false", dest="allow_32bpp", help="Skip 32bpp alternative sprites")
    opt_parser.add_option(
        "--no-palette-validation",
        action="store_true",
        dest="disable_palette_validation",
        help="Disable palette validation for sprites",
    )
    opt_parser.add_option(
        "--list-unused-strings", action="store_true", dest="list_unused_strings", help="List unused strings"
    )

    opts, args = opt_parser.parse_args(argv)

    generic.set_verbosity(0 if opts.quiet else opts.verbosity)
    generic.Warning.disabled = opts.disable_warning
    generic.set_cache_root_dir(None if opts.no_cache else opts.cache_dir)
    spritecache.keep_orphaned = opts.keep_orphaned
    global_constants.allow_extra_zoom = opts.allow_extra_zoom
    global_constants.allow_32bpp = opts.allow_32bpp

    opts.outputfile_given = (
        opts.grf_filename or opts.nfo_filename or opts.nml_filename or opts.dep_filename or opts.outputs
    )

    if not args:
        if not opts.outputfile_given:
            opt_parser.print_help()
            sys.exit(2)
        input_filename = None  # Output filenames, but no input -> use stdin.
    elif len(args) > 1:
        opt_parser.error("Error: only a single nml file can be read per run")
    else:
        input_filename = args[0]
        if not os.access(input_filename, os.R_OK):
            raise generic.ScriptError('Input file "{}" does not exist'.format(input_filename))

    return opts, input_filename


def main(argv):
    global developmode

    opts, input_filename = parse_cli(argv)

    if opts.stack:
        developmode = True

    grfstrings.read_extra_commands(opts.custom_tags)

    generic.print_progress("Reading lang ...")

    grfstrings.read_lang_files(opts.lang_dir, opts.default_lang)

    generic.clear_progress()

    # We have to do the dependency check first or we might later have
    #   more targets than we asked for
    outputs = []
    if opts.dep_check:
        # First make sure we have a file to output the dependencies to:
        dep_filename = opts.dep_filename
        if dep_filename is None and opts.grf_filename is not None:
            dep_filename = filename_output_from_input(opts.grf_filename, ".dep")
        if dep_filename is None and input_filename is not None:
            dep_filename = filename_output_from_input(input_filename, ".dep")
        if dep_filename is None:
            raise generic.ScriptError(
                "-M requires a dependency file either via -MF, an input filename or a valid output via --grf"
            )

        # Now make sure we have a file which is the target for the dependencies:
        depgrf_filename = opts.depgrf_filename
        if depgrf_filename is None and opts.grf_filename is not None:
            depgrf_filename = opts.grf_filename
        if depgrf_filename is None and input_filename is not None:
            depgrf_filename = filename_output_from_input(input_filename, ".grf")
        if depgrf_filename is None:
            raise generic.ScriptError(
                "-M requires either a target grf file via -MT, an input filename or a valid output via --grf"
            )

        # Only append the dependency check to the output targets when we have both,
        #   a target grf and a file to write to
        if dep_filename is not None and depgrf_filename is not None:
            outputs.append(output_dep.OutputDEP(dep_filename, depgrf_filename))

    if input_filename is None:
        input = sys.stdin
    else:
        input = codecs.open(generic.find_file(input_filename), "r", "utf-8")
        # Only append an output grf name, if no output is given, also not implicitly via -M
        if not opts.outputfile_given and not outputs:
            opts.grf_filename = filename_output_from_input(input_filename, ".grf")

    # Translate the 'common' palette names as used by OpenTTD to the traditional ones being within NML's code
    if opts.forced_palette == "DOS":
        opts.forced_palette = "DEFAULT"
    elif opts.forced_palette == "WIN":
        opts.forced_palette = "LEGACY"

    if opts.grf_filename:
        outputs.append(output_grf.OutputGRF(opts.grf_filename))
    if opts.nfo_filename:
        outputs.append(output_nfo.OutputNFO(opts.nfo_filename, opts.start_sprite_num))
    if opts.nml_filename:
        outputs.append(output_nml.OutputNML(opts.nml_filename))

    for output in opts.outputs:
        outroot, outext = os.path.splitext(output)
        outext = outext.lower()
        if outext == ".grf":
            outputs.append(output_grf.OutputGRF(output))
        elif outext == ".nfo":
            outputs.append(output_nfo.OutputNFO(output, opts.start_sprite_num))
        elif outext == ".nml":
            outputs.append(output_nml.OutputNML(output))
        elif outext == ".dep":
            outputs.append(output_dep.OutputDEP(output, opts.grf_filename))
        else:
            generic.print_error("Unknown output format {}".format(outext))
            sys.exit(2)

    ret = nml(
        input,
        input_filename,
        opts.debug,
        outputs,
        opts.start_sprite_num,
        opts.compress,
        opts.crop,
        opts.forced_palette,
        opts.md5_filename,
        opts.debug_parser,
        opts.disable_palette_validation,
    )

    input.close()

    if opts.list_unused_strings:
        grfstrings.list_unused_strings()
    sys.exit(ret)


def filename_output_from_input(name, ext):
    return os.path.splitext(name)[0] + ext


def nml(
    inputfile,
    input_filename,
    output_debug,
    outputfiles,
    start_sprite_num,
    compress_grf,
    crop_sprites,
    forced_palette,
    md5_filename,
    debug_parser,
    disable_palette_validation,
):
    """
    Compile an NML file.

    @param inputfile: File handle associated with the input file.
    @type  inputfile: C{File}

    @param input_filename: Filename of the input file, C{None} if receiving from L{sys.stdin}
    @type  input_filename: C{str} or C{None}

    @param outputfiles: Output streams to write to.
    @type  outputfiles: C{List} of L{output_base.OutputBase}

    @param start_sprite_num: Number of the first sprite.
    @type  start_sprite_num: C{int}

    @param compress_grf: Enable GRF sprite compression.
    @type  compress_grf: C{bool}

    @param crop_sprites: Enable sprite cropping.
    @type  crop_sprites: C{bool}

    @param forced_palette: Palette to use for the file.
    @type  forced_palette: C{str}

    @param md5_filename: Filename to use for writing the md5 sum of the grf file.
                         C{None} if the file should not be written.
    @type  md5_filename: C{str} or C{None}
    """
    generic.OnlyOnce.clear()

    generic.print_progress("Reading {} ...".format(input_filename or "standard input"))

    try:
        script = inputfile.read()
    except UnicodeDecodeError as ex:
        raise generic.ScriptError("Input file is not utf-8 encoded: {}".format(ex))
    # Strip a possible BOM
    script = script.lstrip(str(codecs.BOM_UTF8, "utf-8"))

    if script.strip() == "":
        generic.print_error("Empty input file")
        return 4

    generic.print_progress("Init parser ...")

    nml_parser = parser.NMLParser(debug_parser)
    if input_filename is None:
        input_filename = "input"

    generic.print_progress("Parsing ...")

    result = nml_parser.parse(script, input_filename)
    result.validate([])

    if output_debug > 0:
        result.debug_print(0)

    for outputfile in outputfiles:
        if isinstance(outputfile, output_nml.OutputNML):
            with outputfile:
                outputfile.write(str(result))

    generic.print_progress("Preprocessing ...")

    result.register_names()
    result.pre_process()
    tmp_actions = result.get_action_list()

    generic.print_progress("Generating actions ...")

    actions = []
    for action in tmp_actions:
        if isinstance(action, action1.SpritesetCollection):
            actions.extend(action.get_action_list())
        else:
            actions.append(action)
    actions.extend(action11.get_sound_actions())

    generic.print_progress("Assigning Action2 registers ...")

    action8_index = -1
    for i in range(len(actions) - 1, -1, -1):
        if isinstance(actions[i], (action2var.Action2Var, action2layout.Action2Layout)):
            actions[i].resolve_tmp_storage()
        elif isinstance(actions[i], action8.Action8):
            action8_index = i

    generic.print_progress("Generating strings ...")

    if action8_index != -1:
        lang_actions = []
        # Add plural/gender/case tables
        for lang_pair in grfstrings.langs:
            lang_id, lang = lang_pair
            lang_actions.extend(action0.get_language_translation_tables(lang))
        # Add global strings
        lang_actions.extend(action4.get_global_string_actions())
        actions = actions[: action8_index + 1] + lang_actions + actions[action8_index + 1 :]

    generic.print_progress("Collecting real sprites ...")

    # Collect all sprite files, and put them into buckets of same image and mask files
    sprite_files = {}
    for action in actions:
        if isinstance(action, real_sprite.RealSpriteAction):
            for sprite in action.sprite_list:
                if sprite.is_empty:
                    continue

                if not Image:
                    generic.print_error("PIL (python-imaging) wasn't found, no support for using graphics")
                    sys.exit(3)

                sprite.validate_size()

                file = sprite.file
                if file is not None:
                    file = file.value

                mask_file = sprite.mask_file
                if mask_file is not None:
                    mask_file = mask_file.value

                key = (file, mask_file)
                sprite_files.setdefault(key, []).append(sprite)

    # Check whether we can terminate sprite processing prematurely for
    #     dependency checks
    skip_sprite_processing = True
    for outputfile in outputfiles:
        if isinstance(outputfile, output_dep.OutputDEP):
            with outputfile:
                for f in sprite_files:
                    if f[0] is not None:
                        outputfile.write(f[0])
                    if f[1] is not None:
                        outputfile.write(f[1])
        skip_sprite_processing &= outputfile.skip_sprite_checks()

    if skip_sprite_processing:
        generic.clear_progress()
        return 0

    used_palette = forced_palette

    if not disable_palette_validation:
        generic.print_progress("Checking palette of source images ...")

        last_file = None
        for f_pair in sprite_files:
            # Palette is defined by mask_file, if present. Otherwise, by the main file.
            f = f_pair[1]
            if f is None:
                f = f_pair[0]

            try:
                with Image.open(generic.find_file(f)) as im:
                    # Verify the image is running in Palette mode, if not, skip this file.
                    if im.mode != "P":
                        continue
                    pal = palette.validate_palette(im, f)
            except IOError as ex:
                raise generic.ImageError(str(ex), f)

            if (
                forced_palette != "ANY"
                and pal != forced_palette
                and not (forced_palette == "DEFAULT" and pal == "LEGACY")
            ):
                raise generic.ImageError(
                    "Image has '{}' palette, but you forced the '{}' palette".format(pal, used_palette), f
                )

            if used_palette == "ANY":
                used_palette = pal
            elif pal != used_palette:
                if used_palette in ("LEGACY", "DEFAULT") and pal in ("LEGACY", "DEFAULT"):
                    used_palette = "DEFAULT"
                else:
                    raise generic.ImageError(
                        "Image has '{}' palette, but \"{}\" has the '{}' palette".format(pal, last_file, used_palette),
                        f,
                    )
            last_file = f

    palette_bytes = {"LEGACY": "W", "DEFAULT": "D", "ANY": "A"}
    if used_palette in palette_bytes:
        grf.set_palette_used(palette_bytes[used_palette])
    encoder = None
    for outputfile in outputfiles:
        outputfile.palette = used_palette  # used by RecolourSpriteAction
        if isinstance(outputfile, output_grf.OutputGRF):
            if encoder is None:
                encoder = spriteencoder.SpriteEncoder(compress_grf, crop_sprites, used_palette)
            outputfile.encoder = encoder

    generic.clear_progress()

    # Read all image data, compress, and store in sprite cache
    if encoder is not None:
        encoder.open(sprite_files)

    # If there are any 32bpp sprites hint to openttd that we'd like a 32bpp blitter
    if global_constants.any_32bpp_sprites:
        grf.set_preferred_blitter("3")

    generic.print_progress("Linking actions ...")

    if action8_index != -1:
        actions = [sprite_count.SpriteCountAction(len(actions))] + actions

    for idx, action in enumerate(actions):
        num = start_sprite_num + idx
        action.prepare_output(num)

    # Processing finished, print some statistics
    action0.print_stats()
    actionF.print_stats()
    action7.print_stats()
    action2.print_stats()
    action6.print_stats()
    grf.print_stats()
    global_constants.print_stats()
    action4.print_stats()
    action11.print_stats()

    generic.print_progress("Writing output ...")

    md5 = None
    for outputfile in outputfiles:
        if isinstance(outputfile, output_grf.OutputGRF):
            outputfile.open()
            for action in actions:
                action.write(outputfile)
            outputfile.close()
            md5 = outputfile.get_md5()

        if isinstance(outputfile, output_nfo.OutputNFO):
            outputfile.open()
            for action in actions:
                action.write(outputfile)
            outputfile.close()

    if md5 is not None and md5_filename is not None:
        with open(md5_filename, "w", encoding="utf-8") as f:
            f.write(md5 + "\n")

    if encoder is not None:
        encoder.close()

    generic.clear_progress()
    return 0


def run():
    try:
        main(sys.argv[1:])

    except generic.ScriptError as ex:
        generic.print_error(str(ex))

        if developmode:
            raise  # Reraise exception in developmode
        sys.exit(1)

    except SystemExit:
        raise

    except KeyboardInterrupt:
        generic.print_error("Application forcibly terminated by user.")

        if developmode:
            raise  # Reraise exception in developmode

        sys.exit(1)

    except Exception as ex:  # Other/internal error.
        if developmode:
            raise  # Reraise exception in developmode

        # User mode: print user friendly error message.
        ex_msg = str(ex)
        if len(ex_msg) > 0:
            ex_msg = '"{}"'.format(ex_msg)

        traceback = sys.exc_info()[2]
        # Walk through the traceback object until we get to the point where the exception happened.
        while traceback.tb_next is not None:
            traceback = traceback.tb_next

        lineno = traceback.tb_lineno
        frame = traceback.tb_frame
        code = frame.f_code
        filename = code.co_filename
        name = code.co_name
        del traceback  # Required according to Python docs.

        ex_data = {
            "class": ex.__class__.__name__,
            "version": version,
            "msg": ex_msg,
            "cli": sys.argv,
            "loc": 'File "{}", line {:d}, in {}'.format(filename, lineno, name),
        }

        msg = (
            "nmlc: An internal error has occurred:\n"
            "nmlc-version: {version}\n"
            "Error:    ({class}) {msg}.\n"
            "Command:  {cli}\n"
            "Location: {loc}\n".format(**ex_data)
        )

        generic.print_error(msg)
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    run()



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action0
from nml.ast import base_statement


class CargoTable(base_statement.BaseStatement):
    def __init__(self, cargo_list, pos):
        base_statement.BaseStatement.__init__(self, "cargo table", pos, False, False)
        self.cargo_list = cargo_list

    def register_names(self):
        generic.OnlyOnce.enforce(self, "cargo table")
        for i, cargo in enumerate(self.cargo_list):
            if isinstance(cargo, expression.Identifier):
                self.cargo_list[i] = expression.StringLiteral(cargo.value, cargo.pos)
            expression.parse_string_to_dword(
                self.cargo_list[i]
            )  # we don't care about the result, only validate the input
            if self.cargo_list[i].value in global_constants.cargo_numbers:
                generic.print_warning(
                    generic.Warning.GENERIC,
                    "Duplicate entry in cargo table: {}".format(self.cargo_list[i].value),
                    cargo.pos,
                )
            else:
                global_constants.cargo_numbers[self.cargo_list[i].value] = i

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Cargo table")
        for cargo in self.cargo_list:
            generic.print_dbg(indentation, "Cargo:", cargo.value)

    def get_action_list(self):
        return action0.get_cargolist_action(self.cargo_list)

    def __str__(self):
        ret = "cargotable {\n"
        ret += ", ".join([expression.identifier_to_print(cargo.value) for cargo in self.cargo_list])
        ret += "\n}\n"
        return ret



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, grfstrings
from nml.actions import action6, actionD, base_action


class ActionB(base_action.BaseAction):
    def __init__(self, severity, lang, msg, data, extra_params):
        self.severity = severity
        self.lang = lang
        self.msg = msg
        self.data = data
        self.extra_params = extra_params

    def write(self, file):
        size = 4
        if not isinstance(self.msg, int):
            size += grfstrings.get_string_size(self.msg)
        if self.data is not None:
            size += grfstrings.get_string_size(self.data) + len(self.extra_params)

        file.start_sprite(size)
        file.print_bytex(0x0B)
        self.severity.write(file, 1)
        file.print_bytex(self.lang)
        if isinstance(self.msg, int):
            file.print_bytex(self.msg)
        else:
            file.print_bytex(0xFF)
            file.print_string(self.msg)
        if self.data is not None:
            file.print_string(self.data)
            for param in self.extra_params:
                param.write(file, 1)
        file.newline()
        file.end_sprite()

    def skip_action7(self):
        return False


default_error_msg = {
    "REQUIRES_TTDPATCH": 0,
    "REQUIRES_DOS_WINDOWS": 1,
    "USED_WITH": 2,
    "INVALID_PARAMETER": 3,
    "MUST_LOAD_BEFORE": 4,
    "MUST_LOAD_AFTER": 5,
    "REQUIRES_OPENTTD": 6,
}

error_severity = {
    "NOTICE": 0,
    "WARNING": 1,
    "ERROR": 2,
    "FATAL": 3,
}


def parse_error_block(error):
    action6.free_parameters.save()
    action_list = []
    act6 = action6.Action6()

    severity = actionD.write_action_value(error.severity, action_list, act6, 1, 1)[0]

    langs = [0x7F]
    if isinstance(error.msg, expression.String):
        custom_msg = True
        msg_string = error.msg
        grfstrings.validate_string(msg_string)
        langs.extend(grfstrings.get_translations(msg_string))
        for lang in langs:
            assert lang is not None
    else:
        custom_msg = False
        msg = error.msg.reduce_constant().value

    if error.data is not None:
        error.data = error.data.reduce()
        if isinstance(error.data, expression.String):
            grfstrings.validate_string(error.data)
            langs.extend(grfstrings.get_translations(error.data))
            for lang in langs:
                assert lang is not None
        elif not isinstance(error.data, expression.StringLiteral):
            raise generic.ScriptError(
                "Error parameter 3 'data' should be the identifier of a custom sting", error.data.pos
            )

    params = []
    for expr in error.params:
        if isinstance(expr, expression.Parameter) and isinstance(expr.num, expression.ConstantNumeric):
            params.append(expr.num)
        else:
            tmp_param, tmp_param_actions = actionD.get_tmp_parameter(expr)
            action_list.extend(tmp_param_actions)
            params.append(expression.ConstantNumeric(tmp_param))

    langs = list(set(langs))
    langs.sort()
    for lang in langs:
        if custom_msg:
            msg = grfstrings.get_translation(msg_string, lang)
        if error.data is None:
            data = None
        elif isinstance(error.data, expression.StringLiteral):
            data = error.data.value
        else:
            data = grfstrings.get_translation(error.data, lang)
        if len(act6.modifications) > 0:
            action_list.append(act6)
        action_list.append(ActionB(severity, lang, msg, data, params))

    action6.free_parameters.restore()
    return action_list



__license__ = """
nmlL is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 nmlL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along
with nmlL; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import json
from nml.editors import extract_tables

text1 = """\
{
    "name": "newgrfml",
    "scopeName": "source.newgrfml",
    "fileTypes": [
        ".nml",
        ".pnml"
    ],
    "patterns": [
        {
            "include": "#comment"
        },
        {
            "include": "#numeric-literal"
        },
        {
            "include": "#string"
        },
        {
            "include": "#block"
        },
        {
            "include": "#variable"
        },
        {
            "include": "#feature"
        },
        {
            "include": "#callback"
        }
    ],
    "repository": {
        "comment": {
            "patterns": [
                {
                    "name": "comment.line.newgrfml",
                    "begin": "//",
                    "end": "$"
                },
                {
                    "name": "comment.line.newgrfml",
                    "begin": "#",
                    "end": "$"
                },
                {
                    "name": "comment.block.newgrfml",
                    "begin": "/\\\\*",
                    "end": "\\\\*/"
                }
            ]
        },
        "numeric-literal": {
            "patterns": [
                {
                    "name": "constant.numeric.decimal.newgrfml",
                    "match": "\\\\b[0-9]+\\\\b"
                },
                {
                    "name": "constant.numeric.binary.newgrfml",
                    "match": "\\\\b0(b|B)[01]*\\\\b"
                },
                {
                    "name": "constant.numeric.hex.newgrfml",
                    "match": "\\\\b0(x|X)[0-9a-fA-F]*\\\\b"
                }
            ]
        },
        "string": {
            "patterns": [
                {
                    "name": "string.quoted.double.newgrfml",
                    "begin": "\\"",
                    "beginCaptures": {
                        "0": {
                            "name": "punctuation.definition.string.begin.newgrfml"
                        }
                    },
                    "end": "\\"",
                    "endCaptures": {
                        "0": {
                            "name": "punctuation.definition.string.end.newgrfml"
                        }
                    }
                },
                {
                    "name": "string.quoted.single.newgrfml",
                    "begin": "'",
                    "beginCaptures": {
                        "0": {
                            "name": "punctuation.definition.string.begin.newgrfml"
                        }
                    },
                    "end": "'",
                    "endCaptures": {
                        "0": {
                            "name": "punctuation.definition.string.end.newgrfml"
                        }
                    }
                }
            ]
        },
"""

text2 = """\
        "block": {
            "patterns": [
                {
                    "name": "storage.type.primitive.newgrfml",
                    "match": "blocks"
                }
            ]
        },
"""

text3 = """\
        "variable": {
            "patterns": [
                {
                    "name": "variable.other.property.newgrfml",
                    "match": "variables"
                }
            ]
        },
"""

text4 = """\
        "feature": {
            "patterns": [
                {
                    "name": "support.class.newgrfml",
                    "match": "features"
                }
            ]
        },
"""

text5 = """\
        "callback": {
            "patterns": [
                {
                    "name": "constant.numeric.newgrfml",
                    "match": "callbacks"
                }
            ]
        }
"""

text6 = """\
    }
}
"""


def run():
    line = r"(?<![_$[:alnum:]])(?:(?<=\\.\\.\\.)|(?<!\\.))("
    lineend = r")(?![_$[:alnum:]])(?:(?=\\.\\.\\.)|(?!\\.))"
    with open("newgrfml.tmLanguage.json", "w") as file:
        file.write(text1)
        file.write(text2.replace("blocks", line + "|".join(extract_tables.block_names_table) + lineend))
        file.write(text3.replace("variables", line + "|".join(extract_tables.variables_names_table) + lineend))
        file.write(text4.replace("features", line + "|".join(extract_tables.feature_names_table) + lineend))
        file.write(text5.replace("callbacks", line + "|".join(extract_tables.callback_names_table) + lineend))
        file.write(text6)

    # Simple JSON that contains all keywords for suggestion purposes.
    with open("providers.json", "w") as file:
        output = {}
        output["block"] = extract_tables.block_names_table
        output["variables"] = extract_tables.variables_names_table
        output["features"] = extract_tables.feature_names_table
        output["callbacks"] = extract_tables.callback_names_table
        file.write(json.dumps(output, separators=(",", ":")))



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants, grfstrings
from nml.actions import action8, action14
from nml.ast import base_statement

palette_node = None
blitter_node = None

"""
Statistics about registers used for parameters.
The 1st field is the largest parameter register used.
The 2nd field is the maximum amount of parameter registers available. This is where L{action6.free_parameters} begins.
"""
param_stats = [0, 0x40]


def print_stats():
    """
    Print statistics about used ids.
    """
    if param_stats[0] > 0:
        generic.print_info("GRF parameter registers: {}/{}".format(param_stats[0], param_stats[1]))


def set_palette_used(pal):
    """
    Set the used palette in the action 14 node, if applicable

    @param pal: Palette to use
    @type pal: C{str} of length 1
    """
    if palette_node:
        palette_node.pal = pal


def set_preferred_blitter(blitter):
    """
    Set the preferred blitter in the action14 node, if applicable

    @param blitter: The blitter to set
    @type blitter: C{str} of length 1
    """
    if blitter_node:
        blitter_node.blitter = blitter


class GRF(base_statement.BaseStatement):
    """
    AST Node for a grf block, that supplies (static) information about the GRF
    This is equivalent to actions 8 and 14

    @ivar name: Name of the GRF (short)
    @type name: L{Expression}, should be L{String} else user error

    @ivar desc: Description of the GRF (longer)
    @type name: L{Expression}, should be L{String} else user error

    @ivar grfid: Globally unique identifier of the GRF
    @type grfid: L{Expression}, should be L{StringLiteral} of 4 bytes else user error

    @ivar version: Version of this GRF
    @type version: L{Expression}

    @ivar min_compatible_version: Minimum (older) version of the same GRF that it is compatible with
    @type min_compatible_version: L{Expression}

    @ivar params: List of user-configurable GRF parameters
    @type params: C{list} of L{ParameterDescription}
    """

    def __init__(self, alist, pos):
        base_statement.BaseStatement.__init__(self, "grf-block", pos, False, False)
        self.name = None
        self.desc = None
        self.url = None
        self.grfid = None
        self.version = None
        self.min_compatible_version = None
        self.params = []
        for assignment in alist:
            if isinstance(assignment, ParameterDescription):
                self.params.append(assignment)
            elif assignment.name.value == "name":
                self.name = assignment.value
            elif assignment.name.value == "desc":
                self.desc = assignment.value
            elif assignment.name.value == "url":
                self.url = assignment.value
            elif assignment.name.value == "grfid":
                self.grfid = assignment.value
            elif assignment.name.value == "version":
                self.version = assignment.value
            elif assignment.name.value == "min_compatible_version":
                self.min_compatible_version = assignment.value
            else:
                raise generic.ScriptError("Unknown item in GRF-block: " + str(assignment.name), assignment.name.pos)

    def register_names(self):
        generic.OnlyOnce.enforce(self, "GRF-block")

    def pre_process(self):
        if None in (self.name, self.desc, self.grfid, self.version, self.min_compatible_version):
            raise generic.ScriptError(
                "A GRF-block requires the"
                " 'name', 'desc', 'grfid', 'version' and 'min_compatible_version' properties to be set.",
                self.pos,
            )

        self.grfid = self.grfid.reduce()
        global_constants.constant_numbers["GRFID"] = expression.parse_string_to_dword(self.grfid)
        self.name = self.name.reduce()
        if not isinstance(self.name, expression.String):
            raise generic.ScriptError("GRF-name must be a string", self.name.pos)
        grfstrings.validate_string(self.name)
        self.desc = self.desc.reduce()
        if not isinstance(self.desc, expression.String):
            raise generic.ScriptError("GRF-description must be a string", self.desc.pos)
        grfstrings.validate_string(self.desc)
        if self.url is not None:
            self.url = self.url.reduce()
            if not isinstance(self.url, expression.String):
                raise generic.ScriptError("URL must be a string", self.url.pos)
            grfstrings.validate_string(self.url)
        self.version = self.version.reduce_constant()
        self.min_compatible_version = self.min_compatible_version.reduce_constant()

        param_num = 0
        for param in self.params:
            param.pre_process(expression.ConstantNumeric(param_num))
            param_num = param.num.value + 1
            if param_num > param_stats[1]:
                raise generic.ScriptError(
                    "No free parameters available."
                    " Consider assigning <num> manually and combine multiple bool parameters into"
                    " a single bitmask parameter using <bit>.",
                    self.pos,
                )
            if param_num > param_stats[0]:
                param_stats[0] = param_num

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "GRF")
        generic.print_dbg(indentation + 2, "grfid:")
        self.grfid.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Name:")
        self.name.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Description:")
        self.desc.debug_print(indentation + 4)

        if self.url is not None:
            generic.print_dbg(indentation + 2, "URL:")
            self.url.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Version:")
        self.version.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Minimal compatible version:")
        self.min_compatible_version.debug_print(indentation + 4)

        if len(self.params) > 0:
            generic.print_dbg(indentation + 2, "Params:")
            for param in self.params:
                param.debug_print(indentation + 4)

    def get_action_list(self):
        global palette_node, blitter_node
        palette_node = action14.UsedPaletteNode("A")
        blitter_node = action14.BlitterNode("8")
        action14_root = action14.BranchNode("INFO")
        action14.grf_name_desc_actions(
            action14_root, self.name, self.desc, self.url, self.version, self.min_compatible_version
        )
        action14.param_desc_actions(action14_root, self.params)
        action14_root.subnodes.append(palette_node)
        action14_root.subnodes.append(blitter_node)
        return action14.get_actions(action14_root) + [action8.Action8(self.grfid, self.name, self.desc)]

    def __str__(self):
        ret = "grf {\n"
        ret += "\tgrfid: {};\n".format(str(self.grfid))
        ret += "\tname: {};\n".format(str(self.name))
        ret += "\tdesc: {};\n".format(str(self.desc))
        if self.url is not None:
            ret += "\turl: {};\n".format(self.url)
        ret += "\tversion: {};\n".format(self.version)
        ret += "\tmin_compatible_version: {};\n".format(self.min_compatible_version)
        for param in self.params:
            ret += str(param)
        ret += "}\n"
        return ret


class ParameterSetting:
    def __init__(self, name, value_list):
        self.name = name
        self.value_list = value_list
        self.type = "int"
        self.name_string = None
        self.desc_string = None
        self.min_val = None
        self.max_val = None
        self.def_val = None
        self.bit_num = None
        self.val_names = []
        self.properties_set = set()

    def pre_process(self):
        for set_val in self.value_list:
            self.set_property(set_val.name.value, set_val.value)

    def __str__(self):
        ret = "\t\t{} {{\n".format(self.name)
        for val in self.value_list:
            if val.name.value == "names":
                ret += "\t\t\tnames: {\n"
                for name in val.value.values:
                    ret += "\t\t\t\t{}: {};\n".format(name.name, name.value)
                ret += "\t\t\t};\n"
            else:
                ret += "\t\t\t{}: {};\n".format(val.name, val.value)
        ret += "\t\t}\n"
        return ret

    def debug_print(self, indentation):
        self.name.debug_print(indentation)
        for val in self.value_list:
            val.name.debug_print(indentation + 2)
            if val.name.value == "names":
                for name in val.value.values:
                    name.name.debug_print(indentation + 4)
                    name.value.debug_print(indentation + 4)
            else:
                val.value.debug_print(indentation + 4)

    def set_property(self, name, value):
        """
        Set a single parameter property

        @param name: Name of the property to be set
        @type name: C{str}

        @param value: Value of the property (note: may be an array)
        @type value: L{Expression}
        """
        if name in self.properties_set:
            raise generic.ScriptError(
                "You cannot set the same property twice in a parameter description block", value.pos
            )
        self.properties_set.add(name)
        if name == "names":
            for name_value in value.values:
                num = name_value.name.reduce_constant().value
                desc = name_value.value
                if not isinstance(desc, expression.String):
                    raise generic.ScriptError("setting name description must be a string", desc.pos)
                self.val_names.append((num, desc))
            return
        value = value.reduce(unknown_id_fatal=False)
        if name == "type":
            if not isinstance(value, expression.Identifier) or (value.value != "int" and value.value != "bool"):
                raise generic.ScriptError("setting-type must be either 'int' or 'bool'", value.pos)
            self.type = value.value
        elif name == "name":
            if not isinstance(value, expression.String):
                raise generic.ScriptError("setting-name must be a string", value.pos)
            self.name_string = value
        elif name == "desc":
            if not isinstance(value, expression.String):
                raise generic.ScriptError("setting-description must be a string", value.pos)
            self.desc_string = value
        elif name == "bit":
            if self.type != "bool":
                raise generic.ScriptError("setting-bit is only valid for 'bool' settings", value.pos)
            self.bit_num = value.reduce_constant()
        elif name == "min_value":
            if self.type != "int":
                raise generic.ScriptError("setting-min_value is only valid for 'int' settings", value.pos)
            self.min_val = value.reduce_constant()
        elif name == "max_value":
            if self.type != "int":
                raise generic.ScriptError("setting-max_value is only valid for 'int' settings", value.pos)
            self.max_val = value.reduce_constant()
        elif name == "def_value":
            self.def_val = value.reduce_constant()
            if self.type == "bool" and self.def_val.value != 0 and self.def_val.value != 1:
                raise generic.ScriptError("setting-def_value must be either 0 or 1 for 'bool' settings", value.pos)
        else:
            raise generic.ScriptError("Unknown setting-property " + name, value.pos)


class ParameterDescription:
    free_bits = {}

    def __init__(self, setting_list, num=None, pos=None):
        self.setting_list = setting_list
        self.num = num
        self.pos = pos

    def __str__(self):
        ret = "\tparam"
        if self.num:
            ret += " " + str(self.num)
        ret += " {\n"
        for setting in self.setting_list:
            ret += str(setting)
        ret += "\t}\n"
        return ret

    def debug_print(self, indentation):
        if self.num is not None:
            self.num.debug_print(indentation)
        for setting in self.setting_list:
            setting.debug_print(indentation + 2)

    def pre_process(self, num):
        if self.num is None:
            self.num = num
        self.num = self.num.reduce_constant()
        if self.num.value not in ParameterDescription.free_bits:
            ParameterDescription.free_bits[self.num.value] = list(range(0, 32))
        for setting in self.setting_list:
            setting.pre_process()
        for setting in self.setting_list:
            if setting.type == "int":
                if len(self.setting_list) > 1 or len(ParameterDescription.free_bits[self.num.value]) != 32:
                    raise generic.ScriptError(
                        "When packing multiple settings in one parameter only bool settings are allowed",
                        setting.name.pos,
                    )
                global_constants.settings[setting.name.value] = {"num": self.num.value, "size": 4}
                ParameterDescription.free_bits[self.num.value].clear()
            else:
                if setting.bit_num is None:
                    if len(ParameterDescription.free_bits[self.num.value]) == 0:
                        raise generic.ScriptError("No bits available for this parameter", setting.name.pos)
                    setting.bit_num = expression.ConstantNumeric(ParameterDescription.free_bits[self.num.value].pop(0))
                else:
                    if setting.bit_num.value not in ParameterDescription.free_bits[self.num.value]:
                        raise generic.ScriptError(
                            "Bit {} is already used".format(setting.bit_num.value), setting.name.pos
                        )
                    ParameterDescription.free_bits[self.num.value].remove(setting.bit_num.value)
                global_constants.misc_grf_bits[setting.name.value] = {
                    "param": self.num.value,
                    "bit": setting.bit_num.value,
                }



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic, grfstrings

from .base_expression import ConstantNumeric, Expression, Type
from .string_literal import StringLiteral


class Parameter(Expression):
    def __init__(self, num, pos=None, by_user=False):
        Expression.__init__(self, pos)
        self.num = num
        if by_user and isinstance(num, ConstantNumeric) and not (0 <= num.value <= 63):
            generic.print_warning(
                generic.Warning.GENERIC,
                "Accessing parameters out of the range 0..63 is not supported and may lead to unexpected behaviour.",
                pos,
            )

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Parameter:")
        self.num.debug_print(indentation + 2)

    def __str__(self):
        return "param[{}]".format(self.num)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        num = self.num.reduce(id_dicts)
        if num.type() != Type.INTEGER:
            raise generic.ScriptError("Parameter number must be an integer.", num.pos)
        return Parameter(num, self.pos)

    def supported_by_action2(self, raise_error):
        supported = isinstance(self.num, ConstantNumeric)
        if not supported and raise_error:
            raise generic.ScriptError(
                "Parameter accesses with non-constant numbers are not supported in a switch-block.", self.pos
            )
        return supported

    def supported_by_actionD(self, raise_error):
        return True

    def __eq__(self, other):
        return other is not None and isinstance(other, Parameter) and self.num == other.num

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.num,))


class OtherGRFParameter(Expression):
    def __init__(self, grfid, num, pos=None):
        Expression.__init__(self, pos)
        self.grfid = grfid
        self.num = num

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "OtherGRFParameter:")
        self.grfid.debug_print(indentation + 2)
        self.num.debug_print(indentation + 2)

    def __str__(self):
        return "param[{}, {}]".format(self.grfid, self.num)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        grfid = self.grfid.reduce(id_dicts)
        # Test validity
        parse_string_to_dword(grfid)
        num = self.num.reduce(id_dicts)
        if num.type() != Type.INTEGER:
            raise generic.ScriptError("Parameter number must be an integer.", num.pos)
        return OtherGRFParameter(grfid, num, self.pos)

    def supported_by_action2(self, raise_error):
        if raise_error:
            raise generic.ScriptError(
                "Reading parameters from another GRF is not supported in a switch-block.", self.pos
            )
        return False

    def supported_by_actionD(self, raise_error):
        return True


def parse_string_to_dword(string):
    """
    Convert string literal expression of length 4 to its equivalent 32-bit number.

    @param string: Expression to convert.
    @type  string: L{Expression}

    @return: Value of the converted expression (a 32-bit integer number, little endian).
    @rtype:  C{int}
    """
    if not isinstance(string, StringLiteral) or grfstrings.get_string_size(string.value, False, True) != 4:
        raise generic.ScriptError("Expected a string literal of length 4", string.pos)

    pos = string.pos
    string = string.value
    bytes = []
    i = 0
    try:
        while len(bytes) < 4:
            if string[i] == "\\":
                bytes.append(int(string[i + 1 : i + 3], 16))
                i += 3
            else:
                bytes.append(ord(string[i]))
                i += 1
    except ValueError:
        raise generic.ScriptError("Cannot convert string to integer id", pos)

    return bytes[0] | (bytes[1] << 8) | (bytes[2] << 16) | (bytes[3] << 24)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

# Available units, mapping of unit name to L{Unit} objects.
units = {}


def get_unit(name):
    """
    Get a unit by name.

    @param name: Name of the unit.
    @type  name: C{str}

    @return: The requested unit.
    @rtype:  L{Unit}
    """
    return units[name]


class Unit:
    def __init__(self, name, type, convert, ottd_mul, ottd_shift):
        """
        Unit definition.

        Conversion factor works like this:
        1 reference_unit = convert other_unit
        So nfo_value = property_value / convert * property_specific_conversion_factor
        To avoid using fractions, rational numbers (as 2-tuples) are used instead.

        ottd_mul and ottd_shift are the values taken from OpenTTD's src/strings.cpp and
        are used to calculate the displayed value by OpenTTD. If possible, adjust_values
        increases or decreases the NFO value so that the desired display value is actually
        achieved.

        @ivar name: Name of the unit.
        @type name: C{str}

        @ivar type: Kind of unit.
        @type type: C{str}

        @ivar convert: Conversion factor of the unit.
        @type convert: Either C{int} or a rational tuple (C{int}, C{int})

        @ivar ottd_mul: OpenTTD multiplication factor for displaying a value of this unit.
        @type ottd_mul: C{int}

        @ivar ottd_shift: OpenTTD shift factor for displaying a value of this unit.
        @type ottd_shift: C{int}
        """
        self.name = name
        self.type = type
        self.convert = convert
        self.ottd_mul = ottd_mul
        self.ottd_shift = ottd_shift

    def __str__(self):
        return self.name


def add_unit(name, type, convert, ottd_mul, ottd_shift):
    """
    Construct new unit, and add it to L{units}.

    @param name: Name of the unit.
    @type  name: C{str}

    @param type: Kind of unit.
    @type  type: C{str}

    @param convert: Conversion factor of the unit.
    @type  convert: Either C{int} or a rational tuple (C{int}, C{int})

    @param ottd_mul: OpenTTD multiplication factor for displaying a value of this unit.
    @type  ottd_mul: C{int}

    @param ottd_shift: OpenTTD shift factor for displaying a value of this unit.
    @type  ottd_shift: C{int}
    """
    unit = Unit(name, type, convert, ottd_mul, ottd_shift)
    units[name] = unit


#          name       type     convert      mul  shift
add_unit("nfo", "nfo", 1, 1, 0)  # don't convert, take value literal

# Speed (reference: m/s)
#          name       type     convert      mul  shift
add_unit("mph", "speed", (3125, 1397), 1, 0)
add_unit("km/h", "speed", (18, 5), 103, 6)
add_unit("m/s", "speed", 1, 1831, 12)

# Power (reference: hpI (imperial hp))

#          name       type     convert      mul  shift
add_unit("hp", "power", 1, 1, 0)  # Default to imperial hp
add_unit("kW", "power", (2211, 2965), 6109, 13)
add_unit("hpM", "power", (731, 721), 4153, 12)
add_unit("hpI", "power", 1, 1, 0)

# Weight (reference: ton)

#          name       type     convert      mul  shift
add_unit("ton", "weight", 1, 1, 0)
add_unit("tons", "weight", 1, 1, 0)
add_unit("kg", "weight", 1000, 1000, 0)

# Snowline height

#          name       type     convert      mul  shift
add_unit("snow%", "snowline", (255, 100), 1, 0)
